# -*- coding: utf-8 -*-
"""
WebResearcher Web UI - FastAPI Application Entry
"""
import sys
from pathlib import Path

from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles

sys.path.append('..')
from webui.api.routes import api_router

# 目录配置
BASE_DIR = Path(__file__).resolve().parent
STATIC_DIR = BASE_DIR / "static"
DATA_DIR = BASE_DIR / "data"

# 确保数据目录存在
DATA_DIR.mkdir(parents=True, exist_ok=True)

# 创建 FastAPI 应用
app = FastAPI(
    title="WebResearcher Deep Research Service",
    version="0.2.0",
    description="Web interface for WebResearcher - an intelligent research agent"
)

# CORS 中间件
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# 注册 API 路由
app.include_router(api_router)

# 挂载静态文件
if STATIC_DIR.exists():
    app.mount("/static", StaticFiles(directory=str(STATIC_DIR)), name="static")


@app.get("/", response_class=HTMLResponse)
async def index_page():
    """首页"""
    index_path = STATIC_DIR / "index.html"
    if not index_path.exists():
        raise HTTPException(status_code=404, detail="前端页面未找到")
    return HTMLResponse(index_path.read_text(encoding="utf-8"))


@app.get("/health")
async def health_check():
    """健康检查"""
    return {"status": "ok", "service": "webresearcher-webui"}


if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Prompts
"""
import json
import re
from typing import List, Optional

REACT_SYSTEM_PROMPT = """You are a deep research assistant. Today is {today}. 
Your core function is to conduct thorough, multi-source investigations into any topic. You must handle both broad, open-domain inquiries and queries within specialized academic fields. For every request, synthesize information from credible, diverse sources to deliver a comprehensive, accurate, and objective response. When you have gathered sufficient information and are ready to provide the definitive response, you must enclose the entire final answer within <answer></answer> tags.

# Tools

You may call one or more functions to assist with the user query.

You are provided with function signatures within <tools></tools> XML tags:
<tools>
{tools_text}
</tools>

For each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:
<tool_call>
{{"name": <function-name>, "arguments": <args-json-object>}}
</tool_call>
"""

REACT_SYSTEM_PROMPT_ZH = """你是一个深度研究助手。今天是 {today}。
你的核心功能是对任何主题进行深入、多源的研究调查。你必须处理广泛的开域查询和专业学术领域的查询。对于每个请求，从可信、多样化的来源综合信息，提供全面、准确和客观的回答。当你收集到足够的信息并准备好提供最终答案时，你必须将整个最终答案包含在 <answer></answer> 标签中。

# 工具

你可以调用一个或多个函数来协助处理用户查询。

你将在 <tools></tools> XML 标签中收到函数签名：
<tools>
{tools_text}
</tools>

对于每个函数调用，在 <tool_call></tool_call> XML 标签中返回一个包含函数名称和参数的 json 对象：
<tool_call>
{{"name": <function-name>, "arguments": <args-json-object>}}
</tool_call>
"""

def is_chinese(text: str) -> bool:
    """
    Detect if text contains Chinese characters.
    
    Args:
        text: Input text to check
        
    Returns:
        True if text contains Chinese characters, False otherwise
    """
    if not text:
        return False
    # Check for Chinese characters (CJK Unified Ideographs)
    chinese_pattern = re.compile(r'[\u4e00-\u9fff]+')
    return bool(chinese_pattern.search(text))


TOOL_DESCRIPTIONS = {
    "search": {"type": "function", "function": {"name": "search", "description": "Perform Google web searches then returns a string of the top search results. Accepts multiple queries.max 5 queries", "parameters": {"type": "object", "properties": {"query": {"type": "array", "items": {"type": "string", "description": "The search query."}, "minItems": 1, "description": "The list of search queries."}}, "required": ["query"]}}},
    "visit": {"type": "function", "function": {"name": "visit", "description": "Visit webpage(s) and return the summary of the content.", "parameters": {"type": "object", "properties": {"url": {"type": "array", "items": {"type": "string"}, "description": "The URL(s) of the webpage(s) to visit. Can be a single URL or an array of URLs."}, "goal": {"type": "string", "description": "The specific information goal for visiting webpage(s)."}}, "required": ["url", "goal"]}}},
    "python": {
        "type": "function",
        "function": {
            "name": "python",
            "description": """Executes Python code in a sandboxed environment. Use this tool to perform calculations, data processing, or any Python operations. IMPORTANT: Any output you want to see MUST be printed to standard output using the print() function. Example:{"name": "python", "arguments": {"code": "import numpy as np\\nresult = np.mean([1,2,3])\\nprint(f'The result is: {result}')"}}""",
            "parameters": {
                "type": "object",
                "properties": {
                    "code": {
                        "type": "string",
                        "description": "The Python code to execute. Use \\n for newlines."
                    }
                },
                "required": ["code"]
            }
        }
    },
    "google_scholar": {"type": "function", "function": {"name": "google_scholar", "description": "Leverage Google Scholar to retrieve relevant information from academic publications. Accepts multiple queries. This tool will also return results from google search", "parameters": {"type": "object", "properties": {"query": {"type": "array", "items": {"type": "string", "description": "The search query."}, "minItems": 1, "description": "The list of search queries for Google Scholar."}}, "required": ["query"]}}},
    "parse_file": {"type": "function", "function": {"name": "parse_file", "description": "This is a tool that can be used to parse multiple user uploaded local files such as PDF, DOCX, PPTX, TXT, CSV, XLSX, DOC, ZIP, MP4, MP3.", "parameters": {"type": "object", "properties": {"files": {"type": "array", "items": {"type": "string"}, "description": "The file name of the user uploaded local files to be parsed."}}, "required": ["files"]}}}
}


def _format_tool_desc(tool_item) -> str:
    """Return a JSON string of the tool schema.
    - If `tool_item` is a full schema dict, use it directly.
    - If it's a string and in TOOL_DESCRIPTIONS, use the built-in schema.
    - Otherwise, synthesize a minimal valid schema for custom tools.
    """
    # If a full schema dict is provided, use it directly
    if isinstance(tool_item, dict):
        return json.dumps(tool_item, ensure_ascii=False)
    # Otherwise, treat as tool name string
    tool_name = str(tool_item)
    if tool_name in TOOL_DESCRIPTIONS:
        return json.dumps(TOOL_DESCRIPTIONS[tool_name], ensure_ascii=False)
    # Fallback: synthesize a minimal valid schema for custom tools
    return json.dumps({
        "type": "function",
        "function": {
                "name": tool_name,
                "description": f"Custom tool '{tool_name}' callable by the agent. Provide a JSON 'arguments' object.",
                "parameters": {"type": "object", "properties": {}, "required": []}
            }
        }, ensure_ascii=False)


def get_react_system_prompt_xml(today: str, tools: list, instruction: str = "", question: Optional[str] = None) -> str:
    """
    Generates a system prompt for ReactAgent including descriptions for the specified tools. XML format for tool descriptions.

    Enhancements:
    - Accepts an optional `instruction` that will be appended as a mandatory, task-specific section.
    - Automatically generates minimal OpenAI tool schemas for custom tools not present in TOOL_DESCRIPTIONS.
      Custom tool schema format:
        {"type": "function", "function": {"name": <tool_name>, "description": "Custom tool callable by the agent. Provide a JSON 'arguments' object.", "parameters": {"type": "object", "properties": {}, "required": []}}}
    - If an element in `tools` is already a full tool schema dict, it will be used as-is.
    - Automatically selects Chinese or English prompt based on question language.
    """

    tools_text = "\n".join(_format_tool_desc(tool) for tool in tools)

    # Select prompt based on question language
    use_chinese = question and is_chinese(question)
    base_prompt = REACT_SYSTEM_PROMPT_ZH if use_chinese else REACT_SYSTEM_PROMPT
    prompt = base_prompt.format(today=today, tools_text=tools_text)

    if instruction:
        if use_chinese:
            instruction_text = (
                "\n\n# 任务特定指令\n"
                f"{instruction}\n\n"
                "上述指令是强制性的。在整个对话过程中始终遵循它。"
            )
        else:
            instruction_text = (
                "\n\n# Task-specific Instruction\n"
                f"{instruction}\n\n"
                "The above instruction is mandatory. Always follow it throughout the conversation."
            )
        prompt = prompt + instruction_text

    return prompt

def get_react_system_prompt_fc(today: str, tools: list, instruction: str = "", question: Optional[str] = None) -> str:
    """
    Generates a system prompt for ReactAgent including descriptions for the specified tools. Function Calling format for tool descriptions.
    """
    system_prompt = (
        f"You are a deep research assistant. Today is {today}. "
        "Your core function is to conduct thorough, multi-source investigations. "
        "Use the provided tools to search and visit web pages. "
        "When you have the final answer, respond with the answer directly. Output the answer same language as the user question."
    )
    if instruction:
        system_prompt += f"\n\nTask-specific instruction:\n{instruction}"
    return system_prompt


def get_iterresearch_system_prompt_fc(today: str, instruction: str = "", question: Optional[str] = None) -> str:
    """
    Generate simplified system prompt for IterResearch paradigm in Function Calling mode.
    
    This prompt does NOT include XML format instructions (<tool_call>, <plan>, <report>, etc.),
    allowing the LLM to use native OpenAI-style function calling.
    
    Args:
        today: Current date string
        instruction: Optional custom instruction
        question: Optional question for language detection
        
    Returns:
        System prompt string for Function Calling mode
    """
    instruction_text = ""
    if instruction:
        instruction_text = f"\n\nAdditional persona instructions:\n{instruction}\n"
    
    use_chinese = question and is_chinese(question)
    
    if use_chinese:
        return f"""你是 WebResearcher，一个高级 AI 研究助手。今天是 {today}。
你的目标是通过迭代搜索网络和综合信息，以高准确性和深度回答用户的问题。
{instruction_text}
**核心工作流程：**
你在一个循环中运行。在每一轮中，你将收到原始问题、当前研究报告和上次工具调用的结果。

**你的任务：**
1. 分析当前信息是否足以回答问题
2. 如果需要更多信息，使用提供的工具进行搜索或访问网页
3. 当你有足够信息时，直接提供最终答案

**重要规则：**
- 使用提供的工具来搜索和访问网页
- 当你准备好提供最终答案时，直接回复答案内容，不要再调用工具
- 答案应该全面、准确，并与问题使用相同的语言

**特殊情况：**
- 如果用户只是打招呼（如"你好"），请友好回应并引导用户提出具体问题
"""
    else:
        return f"""You are WebResearcher, an advanced AI research agent. Today is {today}.
Your goal is to answer the user's question with high accuracy and depth by iteratively searching the web and synthesizing information.
{instruction_text}
**Core Workflow:**
You operate in a loop. In each round, you will receive the original question, the current research report, and the result from your last tool call.

**Your Task:**
1. Analyze whether the current information is sufficient to answer the question
2. If more information is needed, use the provided tools to search or visit web pages
3. When you have sufficient information, provide the final answer directly

**Important Rules:**
- Use the provided tools to search and visit web pages
- When you are ready to provide the final answer, respond with the answer content directly without calling tools
- The answer should be comprehensive, accurate, and in the same language as the question

**Special Cases:**
- If the user is just greeting (e.g., "hello"), respond warmly and invite them to ask a specific question
"""


def get_iterresearch_system_prompt(today: str, function_list: list, instruction: str = "", question: Optional[str] = None) -> str:
    """
    Generate system prompt for IterResearch paradigm (XML Protocol mode).
    
    Requires LLM to generate <plan>, <report>, and <tool_call>/<answer> in a single call.
    """
    tools_text = "\n".join(_format_tool_desc(tool) for tool in function_list)
    instruction_text = ""
    if instruction:
        instruction_text = f"\n\nAdditional persona instructions:\n{instruction}\n"
    
    # Select prompt based on question language
    use_chinese = question and is_chinese(question)
    
    if use_chinese:
        ITERRESEARCH_PROMPT = f"""你是 WebResearcher，一个高级 AI 研究助手。
今天是 {today}。你的目标是通过迭代搜索网络和综合信息，以高准确性和深度回答用户的问题。
{instruction_text}

**特殊情况处理：**
- 如果用户只是打招呼（如"你好"、"hi"、"hello"），请友好回应并引导用户提出具体问题。
- 对于简单的社交互动，直接在 <answer> 中提供友好回复，无需调用工具或进行研究。

**IterResearch 核心循环：**
你在一个循环中运行。在每一轮（第 i 轮）中，你将收到原始"问题"、上一轮的"演进报告"（R_{{i-1}}）以及上次工具使用的"观察结果"（O_{{i-1}}）。

你在单次调用中的任务是生成一个包含三个部分的结构化响应，按以下确切顺序：<plan>、<report> 和 <tool_call>（或 <answer> 或 <terminate>）。

**1. `<plan>` 块（认知草稿）：**
   - 首先，分析问题、当前报告（R_{{i-1}}）和最新观察结果（O_{{i-1}}）。
   - 批判性评估：信息是否充足？是否存在空白、矛盾或新线索？
   - 为*当前*轮次制定计划。你现在需要做什么？
   - 这个块是你的私人思考过程，但应该表达为外部计划。
   - 计划应该与问题使用相同的语言。

**2. `<report>` 块（演进中心记忆）：**
   - **关键**，你必须更新你的研究报告（R_i）。
   - 将观察结果（O_{{i-1}}）中的新信息与现有报告（R_{{i-1}}）综合。
   - 这个*新*报告（R_i）应该是*所有*迄今为止发现的全面、精炼和连贯的总结。
   - 它应该纠正任何先前的错误，删除冗余，并整合新事实。
   - 如果观察结果（O_{{i-1}}）没有用或是错误，你仍应说明这一点，并返回*先前*的报告内容不变或进行最小更新。
   - 这个块将是（除了原始问题之外）传递到下一轮的*唯一*记忆。
   - 报告应该与问题使用相同的语言。
   - **对于简单问候**，可以简单说明这是社交互动，无需详细报告。

**3. `<tool_call>`、`<answer>` 或 `<terminate>` 块（行动）：**
   - 基于你的 `<plan>` 和*新更新的* `<report>`，决定下一步。
   - **如果需要更多研究：**
     - 选择一个可用工具。
     - 输出一个包含该工具 JSON 的*单个* `<tool_call>` 块。
   - **如果你有完整和最终答案并想明确呈现：**
     - 不要使用工具。
     - 在 `<answer>` 块内提供最终、全面的答案。
     - 这将终止研究。
   - **如果报告已包含最终答案，你只想停止：**
     - 不要使用工具。
     - 输出 `<terminate>`（可选地在标签内包含简短原因）。
     - 确保 `<report>` 块现在包含与问题相同语言的完整、面向用户的答案。

**输出格式（严格）：**
你的响应*必须*遵循以下确切结构：
<plan>
你对此轮的详细分析和计划。
</plan>
<report>
*新*的、更新的和综合的报告（R_i），整合了最新观察结果。
</report>
<tool_call>
{{"name": "tool_to_use", "arguments": {{"arg1": "value1", ...}}}}
</tool_call>

*或者，如果答案已准备好：*

<plan>
你关于答案完整的推理。
</plan>
<report>
支持答案的最终、完整报告。
</report>
<answer>
用户问题的最终、全面答案。
</answer>

*或者，如果报告已包含最终答案，你准备停止而不重复：*

<plan>
你关于为什么不需要进一步行动或答案的推理。
</plan>
<report>
应该交付给用户的最终、完整报告。与问题使用相同的语言。
</report>
<terminate>
可选：解释停止条件的简短说明。
</terminate>

**可用工具：**
你可以访问以下工具。一次使用一个。
<tools>
{tools_text}
</tools>
"""
    else:
        ITERRESEARCH_PROMPT = f"""You are WebResearcher, an advanced AI research agent. 
Today is {today}. Your goal is to answer the user's question with high accuracy and depth by iteratively searching the web and synthesizing information.
{instruction_text}

**Special Cases Handling:**
- If the user is just greeting (e.g., "hello", "hi", "你好"), respond warmly and invite them to ask a specific question.
- For simple social interactions, provide a friendly response directly in the <answer> block without using tools or conducting research.

**IterResearch Core Loop:**
You operate in a loop. In each round (Round i), you will be given the original "Question", your "Evolving Report" from the previous round (R_{{i-1}}), and the "Observation" from your last tool use (O_{{i-1}}).

Your task in a single turn is to generate a structured response containing three parts in this exact order: <plan>, <report>, and <tool_call> (or <answer> or <terminate>).

**1. `<plan>` Block (Cognitive Scratchpad):**
   - First, analyze the Question, the current Report (R_{{i-1}}), and the latest Observation (O_{{i-1}}).
   - Critically evaluate: Is the information sufficient? Are there gaps, contradictions, or new leads?
   - Formulate a plan for the *current* round. What do you need to do *now*?
   - This block is your private thought process, but should be expressed as an external plan. 
   - The plan should be in the same language as the question.

**2. `<report>` Block (Evolving Central Memory):**
   - **Crucially**, you must update your research report (R_i).
   - Synthesize the new information from the Observation (O_{{i-1}}) with your existing Report (R_{{i-1}}).
   - This *new* report (R_i) should be a comprehensive, refined, and coherent summary of *all* findings so far.
   - It should correct any previous errors, remove redundancies, and integrate new facts.
   - If the observation (O_{{i-1}}) was not useful or was an error, you should still state that and return the *previous* report content unchanged or with minimal updates.
   - This block will be the *only* memory (besides the original question) carried forward to the next round.
   - The report should be in the same language as the question.
   - **For simple greetings**, you can briefly note this is a social interaction without extensive reporting.

**3. `<tool_call>`, `<answer>`, or `<terminate>` Block (Action):**
   - Based on your `<plan>` and your *newly updated* `<report>`, decide the next step.
   - **If more research is needed:**
     - Choose one of the available tools.
     - Output a *single* `<tool_call>` block with the JSON for that tool.
   - **If you have a complete and final answer and want to present it explicitly:**
     - Do NOT use a tool.
     - Provide the final, comprehensive answer inside an `<answer>` block.
     - This will terminate the research.
   - **If the report already contains the finalized answer and you simply want to stop:**
     - Do NOT use a tool.
     - Output `<terminate>` (optionally with a short reason inside the tag).
     - Ensure the `<report>` block now holds the complete, user-facing answer in the same language as the question.

**Output Format (Strict):**
Your response *must* follow this exact structure:
<plan>
Your detailed analysis and plan for this round.
</plan>
<report>
The *new*, updated, and synthesized report (R_i), integrating the latest observation. 
</report>
<tool_call>
{{"name": "tool_to_use", "arguments": {{"arg1": "value1", ...}}}}
</tool_call>

*OR, if the answer is ready:*

<plan>
Your reasoning for why the answer is complete.
</plan>
<report>
The final, complete report that supports the answer.
</report>
<answer>
The final, comprehensive answer to the user's question. 
</answer>

*OR, if the report already contains the final answer and you are ready to stop without repeating it:*

<plan>
Your reasoning for why no further actions or answers are needed.
</plan>
<report>
The final, complete report that should be delivered to the user. Same language as the question.
</report>
<terminate>
Optional: brief note explaining the stop condition.
</terminate>

**Available Tools:**
You have access to the following tools. Use them one at a time.
<tools>
{tools_text}
</tools>
"""
    return ITERRESEARCH_PROMPT


EXTRACTOR_PROMPT = """Please process the following webpage content and user goal to extract relevant information:

## **Webpage Content**
{webpage_content}

## **User Goal**
{goal}

## **Task Guidelines**
1. **Content Scanning for Rational**: Locate the **specific sections/data** directly related to the user's goal within the webpage content
2. **Key Extraction for Evidence**: Identify and extract the **most relevant information** from the content, you never miss any important information, output the **full original context** of the content as far as possible, it can be more than three paragraphs.
3. **Summary Output for Summary**: Organize into a concise paragraph with logical flow, prioritizing clarity and judge the contribution of the information to the goal.

**Final Output Format using JSON format has "rational", "evidence", "summary" feilds**
"""

EXTRACTOR_PROMPT_ZH = """请处理以下网页内容和用户目标以提取相关信息：

## **网页内容**
{webpage_content}

## **用户目标**
{goal}

## **任务指南**
1. **内容扫描（Rational）**：定位网页内容中与用户目标直接相关的**特定部分/数据**
2. **关键提取（Evidence）**：识别并提取内容中**最相关的信息**，你永远不会遗漏任何重要信息，尽可能输出内容的**完整原始上下文**，可以是三个以上的段落。
3. **摘要输出（Summary）**：组织成具有逻辑流程的简洁段落，优先考虑清晰度并判断信息对目标的贡献。

**最终输出格式使用 JSON 格式，包含 "rational"、"evidence"、"summary" 字段**
"""


def get_extractor_prompt(goal: str) -> str:
    """
    Get extractor prompt based on goal language.
    
    Args:
        goal: User goal string
        
    Returns:
        Extractor prompt string in appropriate language
    """
    use_chinese = is_chinese(goal)
    if use_chinese:
        return EXTRACTOR_PROMPT_ZH
    return EXTRACTOR_PROMPT


def get_webweaver_planner_prompt(today: str, tool_list: List[str], instruction: str = "", question: Optional[str] = None) -> str:
    """
    Generate system prompt for WebWeaver Planner Agent.
    
    The Planner explores research questions and produces comprehensive, citation-grounded outlines.
    Based on WebWeaver paper Section 3.2 and Appendix B.2.
    
    Args:
        today: Current date string
        tool_list: List of available tool names
        
    Returns:
        System prompt string for Planner
    """
    tool_list_str = ', '.join(tool_list)
    instruction_text = ""
    if instruction:
        instruction_text = f"\n\nAdditional persona instructions:\n{instruction}\n"
    
    # Select prompt based on question language
    use_chinese = question and is_chinese(question)
    
    if use_chinese:
        return f"""你是 WebWeaver 的规划者智能体。今天是 {today}。你的任务是探索一个研究问题并生成一个全面的、基于引用的提纲。
{instruction_text}

你将把所有发现的证据存储在记忆库中，记忆库会为其分配一个引用 ID。

你在一个 ReAct（计划-行动-观察）循环中运行。
在每一步中，你将收到[问题]、你的[当前提纲]和[最后观察结果]。

你的目标是通过采取以下三种行动之一来迭代完善[当前提纲]：

1.  `<tool_call>`：收集更多信息。
    - 如果[当前提纲]不完整或缺乏证据，请使用此操作。
    - 你有以下工具：{tool_list_str}。
    - 工具将返回新证据的摘要和引用 ID（例如 id_1），该证据现在在记忆库中。
    - 格式：<tool_call>{{"name": "tool_name", "arguments": {{"arg": "value"}}}}</tool_call>

2.  `<write_outline>`：更新或创建研究提纲。
    - 在从工具收集新证据后使用此操作。
    - 你的新提纲*必须*将新的引用 ID（例如 <citation>id_1, id_2</citation>）整合到相关部分。
    - 此操作*替换*下一步的[当前提纲]。
    - **关键：提纲必须与[问题]使用相同的语言编写。**
    - 格式：<write_outline>
1. 引言 <citation>id_1</citation>
 1.1 背景 <citation>id_2</citation>
...
</write_outline>

3.  `<terminate>`：当提纲完整、详细且完全基于引用时。
    - 此操作完成你的工作。
    - 格式：<terminate>

**严格响应格式：**
你必须*仅*使用一个 `<plan>` 块后跟*一个*行动块（`<tool_call>`、`<write_outline>` 或 `<terminate>`）来响应。

示例：
<plan>
你对当前状态的分析以及下一步行动的计划。
</plan>
<tool_call>
{{"name": "search", "arguments": {{"query": ["搜索词1", "搜索词2"]}}}}
</tool_call>

*或者*

<plan>
你对新证据的分析以及如何更新提纲。
</plan>
<write_outline>
新的、完整的、基于引用的提纲。**必须使用与[问题]相同的语言。**
</write_outline>

*或者*

<plan>
提纲已包含所有必要的证据。
</plan>
<terminate>
"""
    else:
        return f"""You are the Planner Agent for WebWeaver. Today is {today}. Your mission is to explore a research question and produce a comprehensive, citation-grounded OUTLINE.
{instruction_text}

You will store all evidence you find in a Memory Bank, which will assign it a citation ID.

You operate in a ReAct (Plan-Action-Observation) loop.
In each step, you will be given the [Question], your [Current Outline], and the [Last Observation].

Your goal is to iteratively refine the [Current Outline] by taking one of three actions:

1.  `<tool_call>`: To gather more information.
    - Use this if the [Current Outline] is incomplete or lacks evidence.
    - You have these tools: {tool_list_str}.
    - The tool will return a summary and a citation ID (e.g., id_1) for the new evidence, which is now in the Memory Bank.
    - Format: <tool_call>{{"name": "tool_name", "arguments": {{"arg": "value"}}}}</tool_call>

2.  `<write_outline>`: To update or create the research outline.
    - Use this after you have gathered new evidence from a tool.
    - Your new outline *must* integrate the new citation IDs (e.g., <citation>id_1, id_2</citation>) into the relevant sections.
    - This action *replaces* the [Current Outline] for the next step.
    - **CRITICAL: The outline MUST be written in the SAME LANGUAGE as the [Question]. If the question is in Chinese, write the outline in Chinese. If in English, write in English.**
    - Format: <write_outline>
1. Introduction <citation>id_1</citation>
 1.1 Background <citation>id_2</citation>
...
</write_outline>

3.  `<terminate>`: When the outline is complete, detailed, and fully citation-grounded.
    - This action finishes your job.
    - Format: <terminate>

**STRICT Response Format:**
You must respond *only* with a `<plan>` block followed by *one* action block (`<tool_call>`, `<write_outline>`, or `<terminate>`).

Example:
<plan>
Your analysis of the current state and your plan for the next action.
</plan>
<tool_call>
{{"name": "search", "arguments": {{"query": ["search term1", "search term2"]}}}}
</tool_call>

*OR*

<plan>
Your analysis of the new evidence and how you will update the outline.
</plan>
<write_outline>
The new, complete, citation-grounded outline. **MUST use the same language as the [Question].**
</write_outline>

*OR*

<plan>
The outline is complete with all necessary evidence.
</plan>
<terminate>
"""


def get_webweaver_writer_prompt(today: str, instruction: str = "", question: Optional[str] = None) -> str:
    """
    Generate system prompt for WebWeaver Writer Agent.
    
    The Writer writes high-quality reports based on the Planner's outline and memory bank.
    Based on WebWeaver paper Section 3.3 and Appendix B.3.
    
    Args:
        today: Current date string
        
    Returns:
        System prompt string for Writer
    """
    instruction_text = ""
    if instruction:
        instruction_text = f"\n\nAdditional persona instructions:\n{instruction}\n"
    
    # Select prompt based on question language
    use_chinese = question and is_chinese(question)
    
    if use_chinese:
        return f"""你是 WebWeaver 的撰写者智能体。今天是 {today}。
你的工作是*仅*基于[最终提纲]和[检索到的证据]撰写高质量、全面的报告。
{instruction_text}

你在一个 ReAct（计划-行动-观察）循环中运行。
你将收到[最终提纲]和[已撰写的报告]。

你的目标是按照提纲逐节撰写报告。

1.  `<plan>`：分析你需要撰写提纲的哪个部分。
    - 查看[最终提纲]和[已撰写的报告]以了解缺少什么。
    - 制定计划。
    - 格式：<plan>...</plan>

2.  `<tool_call>`（行动：`retrieve`）：
    - 基于你的思考，识别*下一*部分所需的引用 ID（例如 "id_1"、"id_2"）。
    - 使用 `retrieve` 工具从记忆库中获取此证据。
    - 格式：<tool_call>{{"name": "retrieve", "arguments": {{"citation_ids": ["id_1", "id_2"]}}}}</tool_call>

3.  `<tool_response>`（观察）：
    - 环境将返回你请求的证据。

4.  `<plan>`：
    - 分析[检索到的证据]。
    - 规划该部分的文本，确保正确使用证据和引用。

5.  `<write>`（行动）：
    - 撰写*当前*部分的完整文本。
    - **关键：报告部分必须与原始[问题]使用相同的语言编写。如果问题是中文，用中文写。如果是英文，用英文写。检查[最终提纲]的语言以确认。**
    - 关键：你*必须*在文本中使用此格式包含原始引用 ID：[cite:id_1]
    - 此文本将追加到[已撰写的报告]。
    - 格式：<write>
## 1.1 引言

这里的文本内容 [cite:id_1]。更多内容 [cite:id_2]。
</write>

6.  `<terminate>`（行动）：
    - 当[最终提纲]的所有部分都已撰写完成时。
    - 格式：<terminate>

**语言要求：**
**整个报告必须与[问题]使用相同的语言。这是强制性的。不要翻译或切换语言。**

**严格响应格式：**
你的响应*必须*遵循计划-行动循环。
- 首先，你*必须*计划，然后 `retrieve`。
- 获得观察结果（证据）后，你*必须*计划，然后 `write`。
- 对所有部分重复此过程。
- 最后，`terminate`。

示例：
<plan>
我需要撰写第 1.1 节。让我检索其证据。
</plan>
<tool_call>
{{"name": "retrieve", "arguments": {{"citation_ids": ["id_1", "id_2"]}}}}
</tool_call>

（观察后）

<plan>
现在我有了证据，我将用与问题相同的语言撰写第 1.1 节。
</plan>
<write>
## 1.1 背景
背景显示... [cite:id_1]。此外... [cite:id_2]。
（必须使用与问题相同的语言）
</write>
"""
    else:
        return f"""You are the Writer Agent for WebWeaver. Today is {today}. 
Your job is to write a high-quality, comprehensive report based *only* on the [Final Outline] and the [Retrieved Evidence].
{instruction_text}

You operate in a ReAct (Plan-Action-Observation) loop.
You will be given the [Final Outline] and the [Report Written So Far].

Your goal is to write the report section by section, following the outline.

1.  `<plan>`: Analyze which section of the outline you need to write next.
    - Look at the [Final Outline] and the [Report Written So Far] to see what's missing.
    - Formulate a plan.
    - Format: <plan>...</plan>

2.  `<tool_call>` (Action: `retrieve`):
    - Based on your thought, identify the citation IDs (e.g., "id_1", "id_2") needed for the *next* section.
    - Use the `retrieve` tool to fetch this evidence from the Memory Bank.
    - Format: <tool_call>{{"name": "retrieve", "arguments": {{"citation_ids": ["id_1", "id_2"]}}}}</tool_call>

3.  `<tool_response>` (Observation):
    - The environment will return the evidence you requested.

4.  `<plan>`:
    - Analyze the [Retrieved Evidence].
    - Plan the prose for the section, making sure to use the evidence and citations correctly.

5.  `<write>` (Action):
    - Write the full text for the *current* section.
    - **CRITICAL: The report section MUST be written in the SAME LANGUAGE as the original [Question].**
    - CRITICAL: You *must* include the original citation IDs in the prose using this format: [cite:id_1]
    - This text will be appended to the [Report Written So Far].
    - Format: <write>
## 1.1 Introduction

Text content here [cite:id_1]. More content [cite:id_2].
</write>

6.  `<terminate>` (Action):
    - When all sections of the [Final Outline] have been written.
    - Format: <terminate>

**LANGUAGE REQUIREMENT:**
**The entire report MUST be in the SAME LANGUAGE as the [Question]. This is MANDATORY. Do NOT translate or switch languages.**

**STRICT Response Format:**
Your response *must* follow the Plan-Action loop.
- First, you *must* Plan, then `retrieve`.
- After you get the Observation (evidence), you *must* Plan, then `write`.
- Repeat this for all sections.
- Finally, `terminate`.

Example:
<plan>
I need to write section 1.1. Let me retrieve the evidence for it.
</plan>
<tool_call>
{{"name": "retrieve", "arguments": {{"citation_ids": ["id_1", "id_2"]}}}}
</tool_call>

(After observation)

<plan>
Now I have the evidence, I'll write section 1.1 in the same language as the question.
</plan>
<write>
## 1.1 Background
The background shows... [cite:id_1]. Furthermore... [cite:id_2].
(MUST use the same language as the question)
</write>
"""



import json
from typing import List, Union
import requests
from webresearcher.base import BaseTool
from openai import OpenAI
import time
import tiktoken
import re
from markdownify import MarkdownConverter
from webresearcher.prompt import get_extractor_prompt
from webresearcher.log import logger
from webresearcher.config import (
    JINA_API_KEY,
    LLM_API_KEY,
    LLM_BASE_URL,
    SUMMARY_MODEL_NAME,
    VISIT_SERVER_MAX_RETRIES,
)


def truncate_to_tokens(text: str, max_tokens: int = 95000) -> str:
    encoding = tiktoken.get_encoding("cl100k_base")
    
    tokens = encoding.encode(text)
    if len(tokens) <= max_tokens:
        return text
    
    truncated_tokens = tokens[:max_tokens]
    return encoding.decode(truncated_tokens)


def extract_readable_text(html: str) -> str:
    """Convert raw HTML into plain text."""
    from bs4 import BeautifulSoup

    soup = BeautifulSoup(html, "html.parser")
    for element in soup(["script", "style", "noscript"]):
        element.decompose()

    text = soup.get_text(separator="\n")
    cleaned_lines = [line.strip() for line in text.splitlines() if line.strip()]
    return "\n".join(cleaned_lines)


def parse_html_to_markdown(html: str, url: str) -> str:
    """Parse HTML to markdown format.
    
    Args:
        html: HTML content to convert
        url: Source URL (used for special handling of Wikipedia pages)
        
    Returns:
        Markdown formatted text
    """
    from bs4 import BeautifulSoup
    
    soup = BeautifulSoup(html, "html.parser")
    title = soup.title.string if soup.title else "No Title"
    
    # Remove javascript, style blocks, and hyperlinks
    for element in soup(["script", "style", "noscript"]):
        element.decompose()
    
    # Remove other common irrelevant elements
    for element in soup.find_all(["nav", "footer", "aside", "form", "figure", "header"]):
        element.decompose()
    
    # Special handling for Wikipedia pages
    if "wikipedia.org" in url:
        body_elm = soup.find("div", {"id": "mw-content-text"})
        title_elm = soup.find("span", {"class": "mw-page-title-main"})
        
        if body_elm:
            main_title = title_elm.string if title_elm else title
            webpage_text = f"# {main_title}\n\n" + MarkdownConverter().convert_soup(body_elm)
        else:
            webpage_text = MarkdownConverter().convert_soup(soup)
    else:
        webpage_text = MarkdownConverter().convert_soup(soup)
    
    # Clean up excessive newlines
    webpage_text = re.sub(r"\r\n", "\n", webpage_text)
    webpage_text = re.sub(r"\n{2,}", "\n\n", webpage_text).strip()
    
    # Add title if not already present
    if not webpage_text.startswith("# "):
        webpage_text = f"# {title}\n\n{webpage_text}"
    
    return webpage_text

OSS_JSON_FORMAT = """# Response Formats
## visit_content
{"properties":{"rational":{"type":"string","description":"Locate the **specific sections/data** directly related to the 
user's goal within the webpage content"},"evidence":{"type":"string","description":"Identify and extract the **most 
relevant information** from the content, never miss any important information, output the **full original context** 
of the content as far as possible, it can be more than three paragraphs.","summary":{"type":"string","description":
"Organize into a concise paragraph with logical flow, prioritizing clarity and judge the contribution of the 
information to the goal."}}}}"""


class Visit(BaseTool):
    # The `description` tells the agent the functionality of this tool.
    name = 'visit'
    description = 'Visit webpage(s) and return the summary of the content.'
    # The `parameters` tell the agent what input parameters the tool has.
    parameters = {
        "type": "object",
        "properties": {
            "url": {
                "type": ["string", "array"],
                "items": {
                    "type": "string"
                    },
                "minItems": 1,
                "description": "The URL(s) of the webpage(s) to visit. Can be a single URL or an array of URLs."
        },
        "goal": {
                "type": "string",
                "description": "The goal of the visit for webpage(s)."
        }
        },
        "required": ["url", "goal"]
    }
    # The `call` method is the main function of the tool.
    def call(self, params: Union[str, dict], **kwargs) -> str:
        try:
            url = params["url"]
            goal = params["goal"]
        except:
            return "[Visit] Invalid request format: Input must be a JSON object containing 'url' and 'goal' fields"

        if isinstance(url, str):
            response = self.readpage_jina(url, goal)
        else:
            response = []
            assert isinstance(url, List)
            start_time = time.time()
            for u in url: 
                if time.time() - start_time > 900:
                    cur_response = "The useful information in {url} for user goal {goal} as follows: \n\n".format(url=url, goal=goal)
                    cur_response += "Evidence in page: \n" + "The provided webpage content could not be accessed. Please check the URL or file format." + "\n\n"
                    cur_response += "Summary: \n" + "The webpage content could not be processed, and therefore, no information is available." + "\n\n"
                else:
                    try:
                        cur_response = self.readpage_jina(u, goal)
                    except Exception as e:
                        cur_response = f"Error fetching {u}: {str(e)}"
                response.append(cur_response)
            response = "\n=======\n".join(response)
        response = response.strip()
        logger.debug(f'[Visit] url: {url},\nSummary Length: {len(response)};\nresponse: {response[:500]}...')
        return response
        
    def call_server(self, msgs, max_retries=2):
        client = OpenAI(
            api_key=LLM_API_KEY,
            base_url=LLM_BASE_URL,
        )
        for attempt in range(max_retries):
            try:
                chat_response = client.chat.completions.create(
                    model=SUMMARY_MODEL_NAME,
                    messages=msgs,
                    temperature=0.7
                )
                content = chat_response.choices[0].message.content
                if content:
                    try:
                        json.loads(content)
                    except:
                        # extract json from string 
                        left = content.find('{')
                        right = content.rfind('}') 
                        if left != -1 and right != -1 and left <= right: 
                            content = content[left:right+1]
                    return content
            except Exception as e:
                logger.debug(f"API call attempt {attempt + 1} failed: {str(e)}")
                if attempt == (max_retries - 1):
                    return ""
                continue


    def jina_readpage(self, url: str) -> str:
        """
        Read webpage content using Jina service.
        
        Args:
            url: The URL to read
            
        Returns:
            str: The webpage content or error message
        """
        # Check if Jina API key is available
        if not JINA_API_KEY or JINA_API_KEY.strip() == "":
            logger.debug(f"[visit] Jina API key not configured, falling back to local fetch")
            return self.local_fetch_url(url)
        
        max_retries = 1  # 减少重试次数从 2 到 1
        timeout = 50
        
        for attempt in range(max_retries):
            headers = {
                "Authorization": f"Bearer {JINA_API_KEY}",
            }
            try:
                response = requests.get(
                    f"https://r.jina.ai/{url}",
                    headers=headers,
                    timeout=timeout
                )
                if response.status_code == 200:
                    webpage_content = response.text
                    return webpage_content
                else:
                    logger.debug(f"Jina API error response: {response.text}")
                    if attempt == max_retries - 1:
                        # 最后一次失败，降级到本地抓取
                        logger.debug(f"[visit] Jina API failed, falling back to local fetch")
                        return self.local_fetch_url(url)
                    raise ValueError("jina readpage error")
            except Exception as e:
                logger.debug(f"[visit] Jina fetch attempt {attempt + 1} failed: {e}")
                time.sleep(0.5)
                if attempt == max_retries - 1:
                    # 最后一次失败，降级到本地抓取
                    logger.debug(f"[visit] Jina API failed, falling back to local fetch")
                    return self.local_fetch_url(url)
                
        return "[visit] Failed to read page."
    
    def local_fetch_url(self, url: str) -> str:
        """
        Fallback method: Fetch URL content locally without Jina.
        
        Args:
            url: The URL to fetch
            
        Returns:
            str: Markdown formatted webpage content or error message
        """
        try:
            headers = {
                "User-Agent": "Mozilla/5.0 (compatible; URLCrawler/1.0; +https://example.com/bot)",
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
                "Accept-Language": "en-US,en;q=0.5",
            }
            
            logger.debug(f"[visit] Local fetching URL: {url}")
            response = requests.get(url, headers=headers, timeout=15, allow_redirects=True)
            response.raise_for_status()
            
            # Check content type
            content_type = response.headers.get("content-type", "").lower()
            allowed_keywords = ("text", "html", "xml")
            if not any(keyword in content_type for keyword in allowed_keywords):
                logger.warning(f"[visit] Unsupported content type: {content_type}")
                return f"[visit] Unsupported content type: {content_type}"
            
            # Decode content
            html_content = response.content.decode(response.encoding or "utf-8", errors="replace")
            
            # Convert to markdown
            markdown_content = parse_html_to_markdown(html_content, url)
            logger.debug(f"[visit] Local fetch successful, content length: {len(markdown_content)}")
            return markdown_content
            
        except requests.exceptions.Timeout:
            logger.warning(f"[visit] Timeout while fetching {url}")
            return f"[visit] Timeout while fetching {url}"
        except requests.exceptions.HTTPError as e:
            logger.warning(f"[visit] HTTP error while fetching {url}: {e}")
            return f"[visit] HTTP error: {e}"
        except Exception as e:
            logger.warning(f"[visit] Failed to fetch {url}: {e}")
            return f"[visit] Failed to fetch page: {e}"

    def html_readpage_jina(self, url: str) -> str:
        max_attempts = 1  # 减少重试次数从 2 到 1
        for attempt in range(max_attempts):
            content = self.jina_readpage(url)
            if content and not content.startswith("[visit] Failed to read page.") and content != "[visit] Empty content." and not content.startswith("[document_parser]"):
                return content
        return "[visit] Failed to read page."

    def readpage_jina(self, url: str, goal: str) -> str:
        """
        Attempt to read webpage content by alternating between jina and aidata services.
        
        Args:
            url: The URL to read
            goal: The goal/purpose of reading the page
            
        Returns:
            str: The webpage content or error message
        """
   
        summary_page_func = self.call_server
        max_retries = VISIT_SERVER_MAX_RETRIES

        content = self.html_readpage_jina(url)

        if content and not content.startswith("[visit] Failed to read page.") and content != "[visit] Empty content." and not content.startswith("[document_parser]"):
            content = truncate_to_tokens(content, max_tokens=95000)
            extractor_prompt_template = get_extractor_prompt(goal)
            messages = [{"role":"user","content": extractor_prompt_template.format(webpage_content=content, goal=goal)}]
            parse_retry_times = 0
            raw = summary_page_func(messages, max_retries=max_retries)
            summary_retries = 1
            while len(raw) < 10 and summary_retries >= 0:
                truncate_length = int(0.7 * len(content)) if summary_retries > 0 else 25000
                status_msg = (
                    f"[visit] Summary url[{url}] " 
                    f"attempt {3 - summary_retries + 1}/3, "
                    f"content length: {len(content)}, "
                    f"truncating to {truncate_length} chars"
                ) if summary_retries > 0 else (
                    f"[visit] Summary url[{url}] failed after 3 attempts, "
                    f"final truncation to 25000 chars"
                )
                logger.debug(status_msg)
                content = content[:truncate_length]
                extractor_prompt_template = get_extractor_prompt(goal)
                extraction_prompt = extractor_prompt_template.format(
                    webpage_content=content,
                    goal=goal
                )
                messages = [{"role": "user", "content": extraction_prompt}]
                raw = summary_page_func(messages, max_retries=max_retries)
                summary_retries -= 1

            parse_retry_times = 1
            if isinstance(raw, str):
                raw = raw.replace("```json", "").replace("```", "").strip()
            while parse_retry_times < 3:
                try:
                    raw = json.loads(raw)
                    break
                except:
                    raw = summary_page_func(messages, max_retries=max_retries)
                    parse_retry_times += 1
            
            if parse_retry_times >= 3:
                useful_information = "The useful information in {url} for user goal {goal} as follows: \n\n".format(url=url, goal=goal)
                useful_information += "Evidence in page: \n" + "The provided webpage content could not be accessed. Please check the URL or file format." + "\n\n"
                useful_information += "Summary: \n" + "The webpage content could not be processed, and therefore, no information is available." + "\n\n"
            else:
                useful_information = "The useful information in {url} for user goal {goal} as follows: \n\n".format(url=url, goal=goal)
                useful_information += "Evidence in page: \n" + str(raw["evidence"]) + "\n\n"
                useful_information += "Summary: \n" + str(raw["summary"]) + "\n\n"

            if len(useful_information) < 10 and summary_retries < 0:
                logger.debug("[visit] Could not generate valid summary after maximum retries")
                useful_information = "[visit] Failed to read page"
            
            return useful_information

        # If no valid content was obtained after all retries
        else:
            useful_information = "The useful information in {url} for user goal {goal} as follows: \n\n".format(url=url, goal=goal)
            useful_information += "Evidence in page: \n" + "The provided webpage content could not be accessed. Please check the URL or file format." + "\n\n"
            useful_information += "Summary: \n" + "The webpage content could not be processed, and therefore, no information is available." + "\n\n"
            return useful_information

# add demo
if __name__ == '__main__':
    logger.debug(f"JINA_API_KEY: {JINA_API_KEY}")
    visit_tool = Visit()
    url = "https://en.wikipedia.org/wiki/Artificial_intelligence"
    goal = "Explain the history of artificial intelligence."
    params = {
        "url": url,
        "goal": goal
    }
    response = visit_tool.call(params)
    logger.debug(f"Response: {response}")



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: WebWeaver Agent - Dual-Agent Research Framework with Dynamic Outline

Supports two tool calling modes:
1. OpenAI Function Calling (default): Uses OpenAI-style tools parameter, works with OpenAI/DeepSeek/etc.
2. XML Protocol: Uses <tool_call> tags, compatible with all LLMs including local models
"""
import json5
import re
import datetime
import asyncio
import random
import time
import json
import inspect

from typing import Any, Callable, Dict, List, Optional, Set
from openai import (
    AsyncOpenAI,
    APIError,
    APIConnectionError,
    APITimeoutError,
    RateLimitError,
    AuthenticationError,
)

from webresearcher.base import BaseTool, today_date
from webresearcher.log import logger
from webresearcher.prompt import get_webweaver_planner_prompt, get_webweaver_writer_prompt, TOOL_DESCRIPTIONS
from webresearcher.tool_memory import MemoryBank, RetrieveTool
from webresearcher.tool_planner_search import PlannerSearchTool
from webresearcher.tool_planner_scholar import PlannerScholarTool
from webresearcher.tool_planner_visit import PlannerVisitTool
from webresearcher.tool_planner_python import PlannerPythonTool
from webresearcher.tool_planner_file import PlannerFileTool
from webresearcher.config import (
    LLM_API_KEY, 
    LLM_BASE_URL, 
    OBS_START, 
    MAX_LLM_CALL_PER_RUN, 
    AGENT_TIMEOUT,
    LLM_MODEL_NAME
)


class BaseWebWeaverAgent:
    """
    Base class for WebWeaver agents (Planner and Writer).
    Handles common LLM calling and tool execution logic.
    
    Supports two tool calling modes:
    - use_xml_protocol=True (default): XML-based <tool_call> tags, works better for structured output
    - use_xml_protocol=False: OpenAI-style function calling
    """

    def __init__(self, llm_config: Dict, tool_map: Dict[str, BaseTool], use_xml_protocol: bool = True):
        """
        Initialize base agent with LLM config and tools.
        
        Args:
            llm_config: LLM configuration dict
            tool_map: Dictionary mapping tool names to BaseTool instances
            use_xml_protocol: If True (default), use XML tags; if False, use OpenAI function calling
        """
        self.llm_config = llm_config
        self.llm_generate_cfg = self.llm_config.get("generate_cfg", {})
        self.model = self.llm_config.get("model", LLM_MODEL_NAME)
        self.llm_timeout = self.llm_config.get("llm_timeout", 300.0)
        self.tool_map = tool_map
        self.function_list = list(tool_map.keys())
        self.api_key = self.llm_config.get("api_key", LLM_API_KEY)
        self.base_url = self.llm_config.get("base_url", LLM_BASE_URL)
        self.use_xml_protocol = use_xml_protocol
        self.client = AsyncOpenAI(
            api_key=self.api_key,
            base_url=self.base_url,
            timeout=self.llm_timeout,
        )
        # Cache for idempotent tool calls to avoid redundant executions
        self.cacheable_tools = set(llm_config.get("cacheable_tools", ["retrieve"]))
        self.tool_call_cache: Dict[str, str] = {}

    def _get_tool_definitions(self) -> List[Dict]:
        """Get tool definitions in OpenAI function calling format."""
        tools = []
        for tool_name in self.function_list:
            if tool_name in TOOL_DESCRIPTIONS:
                tools.append(TOOL_DESCRIPTIONS[tool_name])
            else:
                # Fallback for custom tools
                tools.append({
                    "type": "function",
                    "function": {
                        "name": tool_name,
                        "description": f"Custom tool '{tool_name}'",
                        "parameters": {"type": "object", "properties": {}, "required": []}
                    }
                })
        return tools

    async def call_server(
            self, 
            msgs: List[Dict], 
            stop_sequences: Optional[List[str]] = None,
            max_tries: int = 3,
            tools: Optional[List[Dict]] = None,
    ) -> Dict[str, Any]:
        """
        Async LLM API call with retry logic using AsyncOpenAI.
        Unified return structure matching ReactAgent and WebResearcherAgent.
        
        Args:
            msgs: List of message dicts
            stop_sequences: Optional stop sequences
            max_tries: Maximum retry attempts
            tools: Optional tool definitions for function calling mode
            
        Returns:
            Dict with keys:
            - content: str, the text content
            - reasoning_content: Optional[str], thinking process (for models like DeepSeek R1)
            - tool_calls: Optional[List], native tool calls (when use_xml_protocol=False)
            - raw_message: the original message object
        """
        base_sleep_time = 1
        stop_sequences = stop_sequences or ([OBS_START] if self.use_xml_protocol else None)

        for attempt in range(max_tries):
            try:
                request_params = {
                    "model": self.model,
                    "messages": msgs,
                    "temperature": self.llm_generate_cfg.get('temperature', 0.1),
                    "top_p": self.llm_generate_cfg.get('top_p', 0.95),
                }
                
                # Add stop sequences only for XML mode
                if stop_sequences and self.use_xml_protocol:
                    request_params["stop"] = stop_sequences
                
                # Add tools for function calling mode (non-XML)
                if tools and not self.use_xml_protocol:
                    request_params["tools"] = tools
                
                # Add extra_body for thinking mode (DeepSeek R1 etc.)
                model_thinking_type = self.llm_generate_cfg.get("model_thinking_type", "")
                if model_thinking_type:
                    request_params["extra_body"] = {
                        "thinking": {"type": model_thinking_type}
                    }
                
                # Use native async call
                chat_response = await self.client.chat.completions.create(**request_params)
                
                message = chat_response.choices[0].message
                content = message.content or ""
                
                # Extract reasoning_content if available (DeepSeek R1, etc.)
                reasoning_content = getattr(message, 'reasoning_content', None)
                
                # Extract native tool_calls if available
                tool_calls = getattr(message, 'tool_calls', None)
                
                logger.debug(
                    f"Input messages: {msgs}, \n"
                    f"Reasoning_content: {reasoning_content}, \n"
                    f"Tool_calls: {tool_calls}, \n"
                    f"LLM Response: {content}"
                )
                
                return {
                    "content": content.strip() if content else "",
                    "reasoning_content": reasoning_content,
                    "tool_calls": tool_calls,
                    "raw_message": message,
                }

            except RateLimitError as e:
                logger.warning(f"Attempt {attempt + 1} rate limit error: {e}")
            except AuthenticationError as e:
                logger.error(f"Authentication error: {e}")
                break  # Don't retry auth errors
            except (APIError, APIConnectionError, APITimeoutError) as e:
                logger.warning(
                    f"Attempt {attempt + 1} API error: {e}, base_url: {self.base_url}, model: {self.model}")
            except Exception as e:
                logger.error(f"Attempt {attempt + 1} unexpected error: {e}")

            if attempt < max_tries - 1:
                sleep_time = base_sleep_time * (2 ** attempt) + random.uniform(0, 1)
                sleep_time = min(sleep_time, 30)
                logger.warning(f"Retrying in {sleep_time:.2f}s...")
                await asyncio.sleep(sleep_time)
            else:
                logger.error("All retry attempts exhausted.")

        return {
            "content": "Error: LLM server failed after all retries.",
            "reasoning_content": None,
            "tool_calls": None,
            "raw_message": None,
        }

    async def _execute_function_call(self, tool_call) -> str:
        """Execute an OpenAI-style function call."""
        func_name = tool_call.function.name
        args_str = tool_call.function.arguments
        
        logger.debug(f"Function call: {func_name}({args_str})")
        
        if func_name not in self.tool_map:
            return f"Error: Tool {func_name} not found"
        
        try:
            args = json.loads(args_str) if args_str else {}
        except json.JSONDecodeError:
            return f"Error: Failed to decode arguments: {args_str}"
        
        # Auto-fix common LLM mistakes
        args = self._fix_tool_args(func_name, args)
        
        # Check cache
        cache_key = None
        if func_name in self.cacheable_tools:
            try:
                cache_key = f"{func_name}::" + json.dumps(args, sort_keys=True, ensure_ascii=False)
                if cache_key in self.tool_call_cache:
                    logger.debug(f"Cache hit for tool '{func_name}'")
                    return self.tool_call_cache[cache_key]
            except Exception:
                cache_key = None
        
        tool = self.tool_map[func_name]
        loop = asyncio.get_event_loop()
        
        try:
            if asyncio.iscoroutinefunction(tool.call):
                result = await tool.call(args)
            else:
                result = await loop.run_in_executor(None, tool.call, args)
            
            result_str = str(result) if not isinstance(result, str) else result
            
            # Store in cache
            if cache_key is not None:
                self.tool_call_cache[cache_key] = result_str
            
            return result_str
        except Exception as e:
            logger.error(f"Tool execution failed: {e}")
            return f"Error: Tool execution failed. {e}"

    async def _execute_xml_tool(self, tool_call_str: str) -> str:
        """Execute a tool call from XML <tool_call> block."""
        loop = asyncio.get_event_loop()
        
        try:
            tool_call = json5.loads(tool_call_str)
            tool_name = tool_call.get('name')
            tool_args = tool_call.get('arguments', {})

            if tool_name not in self.tool_map:
                return f"Error: Tool '{tool_name}' not found in agent's tool map."

            # Auto-fix common LLM mistakes
            tool_args = self._fix_tool_args(tool_name, tool_args)

            # Cache check for idempotent tools
            cache_key = None
            if tool_name in self.cacheable_tools:
                try:
                    cache_key = f"{tool_name}::" + json.dumps(tool_args, sort_keys=True, ensure_ascii=False)
                    if cache_key in self.tool_call_cache:
                        logger.debug(f"Cache hit for tool '{tool_name}' with identical arguments.")
                        return self.tool_call_cache[cache_key]
                except Exception:
                    cache_key = None

            tool = self.tool_map[tool_name]

            # Handle async vs sync tools
            if asyncio.iscoroutinefunction(tool.call):
                result = await tool.call(tool_args)
            else:
                result = await loop.run_in_executor(None, tool.call, tool_args)

            result_str = str(result) if not isinstance(result, str) else result

            # Store in cache if applicable
            if cache_key is not None:
                self.tool_call_cache[cache_key] = result_str

            return result_str

        except Exception as e:
            logger.error(f"Tool call failed: {e}")
            return f"Error: Tool call failed. Input: {tool_call_str}. Error: {e}"

    def _fix_tool_args(self, tool_name: str, tool_args: Dict) -> Dict:
        """Auto-fix common LLM mistakes: convert string to array for certain parameters."""
        if tool_name in ['search', 'google_scholar'] and 'query' in tool_args:
            if isinstance(tool_args['query'], str):
                tool_args['query'] = [tool_args['query']]
        elif tool_name == 'visit' and 'url' in tool_args:
            if isinstance(tool_args['url'], str):
                tool_args['url'] = [tool_args['url']]
        elif tool_name == 'parse_file' and 'files' in tool_args:
            if isinstance(tool_args['files'], str):
                tool_args['files'] = [tool_args['files']]
        return tool_args

    # Legacy method for backward compatibility
    async def call_tool(self, tool_call_str: str) -> str:
        """Legacy method - delegates to _execute_xml_tool."""
        return await self._execute_xml_tool(tool_call_str)


class WebWeaverPlanner(BaseWebWeaverAgent):
    """
    Planner Agent for WebWeaver.
    
    Goal: Iteratively search and optimize a research outline.
    Actions: search, write_outline, terminate.
    
    Supports two tool calling modes:
    - use_xml_protocol=True (default): XML-based <tool_call> tags, works better for structured output
    - use_xml_protocol=False: OpenAI-style function calling
    
    Based on WebWeaver paper Section 3.2 and Appendix B.2.
    """

    def __init__(
            self, 
            llm_config: Dict, 
            memory_bank: MemoryBank, 
            function_list: Optional[List[str]] = None, 
            instruction: str = "",
            use_xml_protocol: bool = True,  # XML protocol works better for Planner's structured output
    ):
        """
        Initialize Planner agent.
        
        Args:
            llm_config: LLM configuration
            memory_bank: Shared MemoryBank instance
            function_list: Optional list of tool names to use
            instruction: Optional custom instruction
            use_xml_protocol: If True, use XML tags; if False (default), use OpenAI function calling
        """
        # Planner's tool set - all 5 common tools like WebResearcher
        full_tool_map = {
            "search": PlannerSearchTool(memory_bank),
            "google_scholar": PlannerScholarTool(memory_bank),
            "visit": PlannerVisitTool(memory_bank),
            "python": PlannerPythonTool(memory_bank),
            "parse_file": PlannerFileTool(memory_bank),
        }
        tool_map = {k: v for k, v in full_tool_map.items() if not function_list or k in function_list} or full_tool_map

        super().__init__(llm_config, tool_map, use_xml_protocol=use_xml_protocol)
        self.memory_bank = memory_bank
        self.instruction = instruction
        # system_prompt will be set dynamically in run() based on question language
        self.system_prompt = get_webweaver_planner_prompt(today_date(), self.function_list, instruction)

    def parse_output(self, text: str) -> Dict[str, str]:
        """
        Parse Planner's output: <plan> and (<tool_call> | <write_outline> | <terminate>).
        
        Args:
            text: Raw LLM output
            
        Returns:
            Dict with 'plan', 'action_type', and 'action_content'
        """
        plan_match = re.search(r'<plan>(.*?)</plan>', text, re.DOTALL)
        plan = plan_match.group(1).strip() if plan_match else ""

        tool_call_match = re.search(r'<tool_call>(.*?)</tool_call>', text, re.DOTALL)
        write_outline_match = re.search(r'<write_outline>(.*?)</write_outline>', text, re.DOTALL)
        terminate_match = re.search(r'<terminate>', text, re.DOTALL)

        action_type = None
        action_content = ""

        if terminate_match:
            action_type = "terminate"
        elif write_outline_match:
            action_type = "write_outline"
            action_content = write_outline_match.group(1).strip()
        elif tool_call_match:
            action_type = "tool_call"
            action_content = tool_call_match.group(1).strip()
        else:
            action_type = "error"
            action_content = "No valid action tag found. Must use <tool_call>, <write_outline>, or <terminate>."
            logger.warning(f"Planner output parsing error: {action_content}")

        return {
            "plan": plan,
            "action_type": action_type,
            "action_content": action_content
        }

    async def run(
            self, 
            question: str,
            progress_callback: Optional[Callable[[Dict[str, Any]], Any]] = None,
    ) -> str:
        """
        Execute Planner's research loop.
        
        Args:
            question: Research question
            progress_callback: Optional callback for progress updates
            
        Returns:
            Final outline string
        """
        async def emit(event: Dict[str, Any]):
            if not callable(progress_callback):
                return
            event.setdefault("timestamp", datetime.datetime.utcnow().isoformat())
            event.setdefault("agent", "planner")
            try:
                maybe = progress_callback(event)
                if inspect.isawaitable(maybe):
                    await maybe
            except Exception as callback_err:
                logger.warning(f"progress_callback raised error: {callback_err}")

        logger.debug("--- [WebWeaver] Planner Agent activated ---")
        
        # Update system prompt based on question language
        self.system_prompt = get_webweaver_planner_prompt(
            today_date(), self.function_list, self.instruction, question=question
        )

        current_outline = "Outline is empty. Start by searching for information."
        last_observation = "No observation yet."
        
        # Get tool definitions for function calling mode
        tool_definitions = self._get_tool_definitions() if not self.use_xml_protocol else None

        for i in range(MAX_LLM_CALL_PER_RUN):
            # Build Planner context
            context_str = (
                f"[Question]\n{question}\n\n"
                f"[Current Outline]\n{current_outline}\n\n"
                f"[Last Observation]\n{last_observation}\n\n"
                f"**IMPORTANT: When you write the outline using <write_outline>, "
                f"you MUST use the SAME LANGUAGE as the [Question] above. Do NOT translate.**"
            )
            # Final iteration: force LLM to output <write_outline>
            is_last_iteration = (i == MAX_LLM_CALL_PER_RUN - 1)
            if is_last_iteration:
                context_str += (
                    "\n[Final Instruction]\n"
                    "This is your last allowed step. You MUST output <write_outline> with the complete final outline. "
                    "Do NOT output <tool_call> or <terminate>."
                )
            messages = [
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": context_str}
            ]

            # Call LLM
            response = await self.call_server(messages, tools=tool_definitions)
            content = response["content"]
            reasoning_content = response["reasoning_content"]
            tool_calls = response["tool_calls"]
            
            # Log and emit reasoning content if available
            if reasoning_content:
                logger.info(f"[Planner Step {i + 1}] Thinking: {reasoning_content[:500]}...")
                await emit({"type": "thinking", "step": i + 1, "content": reasoning_content})

            # === Function Calling Mode: Handle native tool calls ===
            if not self.use_xml_protocol and tool_calls:
                for tool_call in tool_calls:
                    tool_result = await self._execute_function_call(tool_call)
                    logger.debug(f"Planner Step {i + 1}: Tool {tool_call.function.name} executed.")
                    
                    await emit({
                        "type": "tool",
                        "step": i + 1,
                        "tool_name": tool_call.function.name,
                        "tool_args": tool_call.function.arguments,
                        "observation": tool_result[:1000],
                    })
                    last_observation = tool_result
                
                # Parse any outline from content
                if content:
                    parsed = self.parse_output(content)
                    if parsed['action_type'] == "write_outline":
                        current_outline = parsed['action_content']
                        await emit({"type": "outline_updated", "step": i + 1, "outline": current_outline})
                    elif parsed['action_type'] == "terminate":
                        logger.debug("Planner finished. Terminating.")
                        return current_outline
                continue

            # === XML Protocol Mode: Parse structured output ===
            parsed = self.parse_output(content)
            logger.debug(f"Planner Step {i + 1} | Action: {parsed}")
            
            await emit({
                "type": "step",
                "step": i + 1,
                "plan": parsed["plan"],
                "action_type": parsed["action_type"],
                "reasoning_content": reasoning_content,
            })

            # Execute action
            if parsed['action_type'] == "terminate":
                logger.debug("Planner finished. Terminating.")
                return current_outline

            elif parsed['action_type'] == "write_outline":
                current_outline = parsed['action_content']
                last_observation = "Outline successfully updated."
                logger.debug(f"Planner Step {i + 1}: Outline updated.")
                await emit({"type": "outline_updated", "step": i + 1, "outline": current_outline})

            elif parsed['action_type'] == "tool_call":
                tool_call_str = parsed['action_content']
                last_observation = await self._execute_xml_tool(tool_call_str)
                logger.debug(f"Planner Step {i + 1}: Tool executed.")
                await emit({
                    "type": "tool",
                    "step": i + 1,
                    "tool_call": tool_call_str,
                    "observation": last_observation[:1000],
                })

            elif parsed['action_type'] == "error":
                last_observation = parsed['action_content']
                logger.warning(f"Planner Step {i + 1}: Action parse error.")

        logger.warning("Planner reached max iterations.")
        return current_outline


class WebWeaverWriter(BaseWebWeaverAgent):
    """
    Writer Agent for WebWeaver.
    
    Goal: Write report section-by-section based on Planner's outline and memory bank.
    Actions: retrieve, write, terminate.
    
    Supports two tool calling modes:
    - use_xml_protocol=True (default): XML-based <tool_call> tags, works better for structured output
    - use_xml_protocol=False: OpenAI-style function calling
    
    Based on WebWeaver paper Section 3.3 and Appendix B.3.
    """

    def __init__(
            self, 
            llm_config: Dict, 
            memory_bank: MemoryBank, 
            instruction: str = "",
            use_xml_protocol: bool = True,  # XML protocol works better for Writer's structured output
    ):
        """
        Initialize Writer agent.
        
        Args:
            llm_config: LLM configuration
            memory_bank: Shared MemoryBank instance
            instruction: Optional custom instruction
            use_xml_protocol: If True, use XML tags; if False (default), use OpenAI function calling
        """
        # Writer only needs retrieve tool
        tool_map = {
            "retrieve": RetrieveTool(memory_bank),
        }

        super().__init__(llm_config, tool_map, use_xml_protocol=use_xml_protocol)
        self.memory_bank = memory_bank
        self.instruction = instruction
        # system_prompt will be set dynamically in run() based on question language
        self.system_prompt = get_webweaver_writer_prompt(today_date(), instruction)

    def parse_output(self, text: str) -> Dict[str, str]:
        """
        Parse Writer's output: <plan> and (<tool_call> | <write> | <terminate>).
        
        Args:
            text: Raw LLM output
            
        Returns:
            Dict with 'plan', 'action_type', and 'action_content'
        """
        plan_match = re.search(r'<plan>(.*?)</plan>', text, re.DOTALL)
        plan = plan_match.group(1).strip() if plan_match else ""

        tool_call_match = re.search(r'<tool_call>(.*?)</tool_call>', text, re.DOTALL)
        write_match = re.search(r'<write>(.*?)</write>', text, re.DOTALL)
        terminate_match = re.search(r'<terminate>', text, re.DOTALL)

        action_type = None
        action_content = ""

        if terminate_match:
            action_type = "terminate"
        elif write_match:
            action_type = "write"
            action_content = write_match.group(1).strip()
        elif tool_call_match:
            action_type = "tool_call"
            action_content = tool_call_match.group(1).strip()
        else:
            action_type = "error"
            action_content = "No valid action tag found. Must use <tool_call> (retrieve), <write>, or <terminate>."
            logger.warning(f"Writer output parsing error: {action_content}")

        return {
            "plan": plan,
            "action_type": action_type,
            "action_content": action_content
        }

    async def run(
            self, 
            question: str, 
            final_outline: str,
            progress_callback: Optional[Callable[[Dict[str, Any]], Any]] = None,
    ) -> str:
        """
        Execute Writer's writing loop.
        
        Args:
            question: Research question
            final_outline: Final outline from Planner
            progress_callback: Optional callback for progress updates
            
        Returns:
            Final report string
        """
        async def emit(event: Dict[str, Any]):
            if not callable(progress_callback):
                return
            event.setdefault("timestamp", datetime.datetime.utcnow().isoformat())
            event.setdefault("agent", "writer")
            try:
                maybe = progress_callback(event)
                if inspect.isawaitable(maybe):
                    await maybe
            except Exception as callback_err:
                logger.warning(f"progress_callback raised error: {callback_err}")

        logger.debug("--- [WebWeaver] Writer Agent activated ---")
        
        # Update system prompt based on question language
        self.system_prompt = get_webweaver_writer_prompt(today_date(), self.instruction, question=question)

        report_written_so_far = ""
        last_observation = "No observation yet. Start by retrieving evidence for the first section."
        # Track retrieve calls to avoid redundant tool executions
        seen_retrieve_keys: Set[str] = set()
        retrieve_repeat_counts: Dict[str, int] = {}
        retrieve_results_cache: Dict[str, str] = {}
        steps_since_last_write = 0
        MAX_IDLE_BEFORE_FORCE_WRITE_HINT = 6
        
        # Get tool definitions for function calling mode
        tool_definitions = self._get_tool_definitions() if not self.use_xml_protocol else None

        for i in range(MAX_LLM_CALL_PER_RUN):
            # Build Writer context
            context_str = (
                f"[Question]\n{question}\n\n"
                f"[Final Outline]\n{final_outline}\n\n"
                f"[Report Written So Far]\n{report_written_so_far}\n\n"
                f"[Last Observation]\n{last_observation}\n\n"
                f"**CRITICAL LANGUAGE REQUIREMENT: The report you write using <write> MUST be "
                f"in the SAME LANGUAGE as the [Question] and [Final Outline] above. "
                f"Check the language carefully and DO NOT translate or switch languages.**"
            )
            # Final iteration: force LLM to output <write>
            is_last_iteration = (i == MAX_LLM_CALL_PER_RUN - 1)
            if is_last_iteration:
                context_str += (
                    "\n[Final Instruction]\n"
                    "This is your last allowed step. You MUST output <write> with a well-structured final section. "
                    "Do NOT output <tool_call> or <terminate>."
                )
            messages = [
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": context_str}
            ]

            # Call LLM
            response = await self.call_server(messages, tools=tool_definitions)
            content = response["content"]
            reasoning_content = response["reasoning_content"]
            tool_calls = response["tool_calls"]
            
            # Log and emit reasoning content if available
            if reasoning_content:
                logger.info(f"[Writer Step {i + 1}] Thinking: {reasoning_content[:500]}...")
                await emit({"type": "thinking", "step": i + 1, "content": reasoning_content})

            # === Function Calling Mode: Handle native tool calls ===
            if not self.use_xml_protocol and tool_calls:
                for tool_call in tool_calls:
                    func_name = tool_call.function.name
                    args_str = tool_call.function.arguments
                    
                    # Check for duplicate retrieve calls
                    if func_name == "retrieve":
                        try:
                            args = json.loads(args_str) if args_str else {}
                            key = json.dumps(args, sort_keys=True, ensure_ascii=False)
                        except Exception:
                            key = None
                        
                        if key and key in seen_retrieve_keys:
                            retrieve_repeat_counts[key] = retrieve_repeat_counts.get(key, 1) + 1
                            cached_evidence = retrieve_results_cache.get(key, "")
                            last_observation = (
                                f"Evidence already retrieved. Here it is again:\n\n{cached_evidence}\n\n"
                                "You MUST now proceed to <write> the section."
                            )
                            steps_since_last_write += 1
                            continue
                        elif key:
                            seen_retrieve_keys.add(key)
                    
                    tool_result = await self._execute_function_call(tool_call)
                    logger.debug(f"Writer Step {i + 1}: Tool {func_name} executed.")
                    
                    # Cache retrieve results
                    if func_name == "retrieve" and key:
                        retrieve_results_cache[key] = tool_result
                    
                    await emit({
                        "type": "tool",
                        "step": i + 1,
                        "tool_name": func_name,
                        "tool_args": args_str,
                        "observation": tool_result[:1000],
                    })
                    last_observation = tool_result
                    steps_since_last_write += 1
                
                # Parse any write/terminate from content
                if content:
                    parsed = self.parse_output(content)
                    if parsed['action_type'] == "write":
                        section_prose = parsed['action_content']
                        report_written_so_far += "\n\n" + section_prose
                        last_observation = f"Section written successfully:\n{section_prose}\n"
                        steps_since_last_write = 0
                        await emit({"type": "section_written", "step": i + 1, "content": section_prose})
                    elif parsed['action_type'] == "terminate":
                        logger.debug("Writer finished. Terminating.")
                        return report_written_so_far
                
                # Force write hint if idling too long
                if steps_since_last_write >= MAX_IDLE_BEFORE_FORCE_WRITE_HINT:
                    last_observation += "\nInstruction: You MUST output <write> now."
                continue

            # === XML Protocol Mode: Parse structured output ===
            parsed = self.parse_output(content)
            logger.debug(f"Writer Step {i + 1} | Action: {parsed}")
            
            await emit({
                "type": "step",
                "step": i + 1,
                "plan": parsed["plan"],
                "action_type": parsed["action_type"],
                "reasoning_content": reasoning_content,
            })

            # Execute action
            if parsed['action_type'] == "terminate":
                logger.debug("Writer finished. Terminating.")
                return report_written_so_far

            elif parsed['action_type'] == "write":
                section_prose = parsed['action_content']
                report_written_so_far += "\n\n" + section_prose
                last_observation = f"Section written successfully:\n{section_prose}\n"
                logger.debug(f"Writer Step {i + 1}: Section written.")
                steps_since_last_write = 0
                await emit({"type": "section_written", "step": i + 1, "content": section_prose})

            elif parsed['action_type'] == "tool_call":
                tool_call_str = parsed['action_content']
                # Try to parse tool call for caching
                try:
                    tool_call_parsed = json5.loads(tool_call_str)
                    tool_name = tool_call_parsed.get('name')
                    tool_args = tool_call_parsed.get('arguments', {})
                except Exception:
                    tool_name, tool_args = None, None

                # Check for duplicate retrieve calls
                if tool_name == "retrieve":
                    try:
                        key = json.dumps(tool_args, sort_keys=True, ensure_ascii=False)
                    except Exception:
                        key = None

                    if key and key in seen_retrieve_keys:
                        retrieve_repeat_counts[key] = retrieve_repeat_counts.get(key, 1) + 1
                        cached_evidence = retrieve_results_cache.get(key, "")
                        last_observation = (
                            f"Evidence already retrieved:\n\n{cached_evidence}\n\n"
                            "You MUST now proceed to <write> the section."
                        )
                        steps_since_last_write += 1
                        continue
                    elif key:
                        seen_retrieve_keys.add(key)

                last_observation = await self._execute_xml_tool(tool_call_str)
                
                # Cache retrieve results
                if tool_name == "retrieve" and key:
                    retrieve_results_cache[key] = last_observation
                
                logger.debug(f"Writer Step {i + 1}: Evidence retrieved.")
                steps_since_last_write += 1
                await emit({
                    "type": "tool",
                    "step": i + 1,
                    "tool_call": tool_call_str,
                    "observation": last_observation[:1000],
                })

            elif parsed['action_type'] == "error":
                last_observation = parsed['action_content']
                logger.warning(f"Writer Step {i + 1}: Action parse error.")
                steps_since_last_write += 1

            # Force write hint if idling too long
            if steps_since_last_write >= MAX_IDLE_BEFORE_FORCE_WRITE_HINT:
                last_observation += (
                    "\nInstruction: You have gathered sufficient evidence. "
                    "You MUST output <write> with a well-structured section now."
                )

        logger.warning("Writer reached max iterations.")
        return report_written_so_far


class WebWeaverAgent:
    """
    WebWeaver main orchestrator.
    
    Coordinates Planner and Writer agents in sequence.
    Based on WebWeaver paper dual-agent framework.
    
    Supports two tool calling modes:
    - use_xml_protocol=True (default): XML-based <tool_call> tags, works better for structured output
    - use_xml_protocol=False: OpenAI-style function calling
    """

    def __init__(
            self,
            llm_config: Optional[Dict] = None,
            function_list: Optional[List[str]] = None,
            instruction: str = "",
            api_key: Optional[str] = None,
            base_url: Optional[str] = None,
            model: Optional[str] = None,
            use_xml_protocol: bool = True,  # XML protocol works better for WebWeaver's structured output
    ):
        """
        Initialize WebWeaver agent.
        
        Args:
            llm_config: LLM configuration dict
            function_list: Optional list of tool names for Planner
            instruction: Optional custom instruction
            api_key: Optional API key override
            base_url: Optional base URL override
            model: Optional model name override
            use_xml_protocol: If True, use XML tags; if False (default), use OpenAI function calling
        """
        base_config = llm_config or {
            "model": LLM_MODEL_NAME,
            "generate_cfg": {
                "temperature": 0.1,
                "top_p": 0.95,
                "max_tokens": 10000,
            },
            "llm_timeout": 300.0,
        }
        self.llm_config = dict(base_config)
        if api_key:
            self.llm_config["api_key"] = api_key
        if base_url:
            self.llm_config["base_url"] = base_url
        if model:
            self.llm_config["model"] = model
        
        self.use_xml_protocol = use_xml_protocol
        
        # Initialize shared memory bank
        self.memory_bank = MemoryBank()

        # Initialize sub-agents with use_xml_protocol
        self.planner = WebWeaverPlanner(
            self.llm_config, 
            self.memory_bank, 
            function_list=function_list, 
            instruction=instruction,
            use_xml_protocol=use_xml_protocol,
        )
        self.writer = WebWeaverWriter(
            self.llm_config, 
            self.memory_bank, 
            instruction=instruction,
            use_xml_protocol=use_xml_protocol,
        )

        logger.debug("WebWeaver Dual-Agent Framework initialized.")
        logger.debug(f"Planner Tools: {self.planner.function_list}")
        logger.debug(f"Writer Tools: {self.writer.function_list}")
        logger.debug(f"Tool Calling Mode: {'XML Protocol' if use_xml_protocol else 'Function Calling'}")

    async def run(
            self, 
            question: str,
            progress_callback: Optional[Callable[[Dict[str, Any]], Any]] = None,
    ) -> Dict[str, Any]:
        """
        Execute WebWeaver's complete dual-agent workflow.
        
        Args:
            question: Research question
            progress_callback: Optional callback for progress updates
            
        Returns:
            Dict with final_report, final_outline, and metadata
        """
        async def emit(event: Dict[str, Any]):
            if not callable(progress_callback):
                return
            event.setdefault("timestamp", datetime.datetime.utcnow().isoformat())
            try:
                maybe = progress_callback(event)
                if inspect.isawaitable(maybe):
                    await maybe
            except Exception as callback_err:
                logger.warning(f"progress_callback raised error: {callback_err}")

        start_time = time.time()
        
        await emit({"type": "status", "status": "starting", "phase": "planner"})

        # Phase 1: Run Planner
        try:
            final_outline = await asyncio.wait_for(
                self.planner.run(question, progress_callback=progress_callback),
                timeout=AGENT_TIMEOUT
            )
            logger.debug("--- Planner Phase Complete ---")
            logger.debug(f"Final Outline:\n{final_outline}")
            logger.debug(f"Memory Bank contains {self.memory_bank.size()} items.")
            
            await emit({
                "type": "phase_complete", 
                "phase": "planner",
                "outline": final_outline,
                "memory_bank_size": self.memory_bank.size(),
            })
        except Exception as e:
            logger.error(f"Planner Agent failed: {e}")
            await emit({"type": "error", "phase": "planner", "error": str(e)})
            return {
                "question": question,
                "final_report": "",
                "error": f"Planner phase error: {e}"
            }

        await emit({"type": "status", "status": "starting", "phase": "writer"})

        # Phase 2: Run Writer
        try:
            final_report = await asyncio.wait_for(
                self.writer.run(question, final_outline, progress_callback=progress_callback),
                timeout=AGENT_TIMEOUT
            )
            logger.debug("--- Writer Phase Complete ---")
            
            await emit({
                "type": "phase_complete",
                "phase": "writer",
                "report": final_report,
            })
        except Exception as e:
            logger.error(f"Writer Agent failed: {e}")
            await emit({"type": "error", "phase": "writer", "error": str(e)})
            return {
                "question": question,
                "final_report": "",
                "error": f"Writer phase error: {e}"
            }

        end_time = time.time()
        
        result = {
            "question": question,
            "final_outline": final_outline,
            "final_report": final_report,
            "memory_bank_size": self.memory_bank.size(),
            "total_time_seconds": end_time - start_time
        }
        
        await emit({
            "type": "complete",
            "result": result,
        })

        return result


async def main():
    """Main function for testing WebWeaver agent."""
    from webresearcher.log import set_log_level
    set_log_level("DEBUG")
    
    llm_config = {
        "model": LLM_MODEL_NAME,
        "generate_cfg": {
            "temperature": 0.1,
            "top_p": 0.95,
            "max_tokens": 10000,
        },
        "llm_timeout": 300.0,
    }
    agent = WebWeaverAgent(llm_config=llm_config)
    print(LLM_API_KEY)
    question =  "刘翔破纪录时候是多少岁?"
    result = await agent.run(question)
    print(result)


if __name__ == '__main__':
    asyncio.run(main())



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: WebResearcher Agent implementing the IterResearch paradigm.

Supports two tool calling modes:
1. OpenAI Function Calling (default): Uses OpenAI-style tools parameter, works with OpenAI/DeepSeek/etc.
2. XML Protocol: Uses <tool_call> tags, compatible with all LLMs including local models
"""
import json
import json5
import re
import datetime
import asyncio
import random
import time

from typing import Any, Callable, Dict, List, Optional
from openai import (
    AsyncOpenAI,
    APIError,
    APIConnectionError,
    APITimeoutError,
    RateLimitError,
    AuthenticationError,
)
import inspect
import sys
sys.path.append('..')
from webresearcher.base import Message, today_date, build_text_completion_prompt, count_tokens as count_tokens_base
from webresearcher.log import logger
from webresearcher.prompt import get_iterresearch_system_prompt, get_iterresearch_system_prompt_fc, TOOL_DESCRIPTIONS
from webresearcher.tool_file import FileParser
from webresearcher.tool_scholar import Scholar
from webresearcher.tool_python import PythonInterpreter
from webresearcher.tool_search import Search
from webresearcher.tool_visit import Visit
from webresearcher.config import (
    LLM_API_KEY, 
    LLM_BASE_URL, 
    OBS_START, 
    OBS_END, 
    MAX_LLM_CALL_PER_RUN,
    FILE_DIR,
    LLM_MODEL_NAME
)


TOOL_CLASS = [
    FileParser(),
    Scholar(),
    Visit(),
    Search(),
    PythonInterpreter(),
]
TOOL_MAP = {tool.name: tool for tool in TOOL_CLASS}


class ResearchRound:
    """
    实现了 IterResearch 范式的核心状态管理器。
    
    状态只包含用于构建下一个提示的核心信息：
    - question (Q): 原始问题
    - current_report (R_{i-1}): 上一轮生成的报告
    - last_observation (O_{i-1}): 上一轮工具调用的结果
    """

    def __init__(self, question: str):
        self.question = question

        # R_{i-1}: 上一轮生成的报告，初始为空
        self.current_report = "This is the first round. The report is empty."

        # O_{i-1}: 上一轮工具调用的结果，初始为空
        self.last_observation = "This is the first round. No tool has been called yet."

    def get_context(self, system_prompt: str) -> List[Dict]:
        """
        构建当前轮次的精简上下文。
        这代表了论文中的 Workspace，即状态 s_t = (Q, R_{i-1}, O_{i-1})
        """
        user_content = (
            f"**Question:** {self.question}\n\n"
            f"**Current Report (R_{{i-1}}):**\n{self.current_report}\n\n"
            f"**Last Observation (O_{{i-1}}):**\n{self.last_observation}"
        )

        return [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_content}
        ]


class WebResearcherAgent:
    """
    Web Research Agent implementing the IterResearch paradigm.
    
    The IterResearch paradigm uses a cyclical process of:
    1. Workspace (question + previous report + last tool result)
    2. Plan-Action (LLM generates plan and action)
    3. Tool Execution (execute the tool and get result)
    4. Synthesis (update report by integrating new information)
    
    Supports two tool calling modes:
    - use_xml_protocol=True (default): XML-based <tool_call> tags, works better for IterResearch paradigm
    - use_xml_protocol=False: OpenAI-style function calling, works with OpenAI/DeepSeek/etc.
    """

    def __init__(
            self,
            llm_config: Optional[Dict] = None,
            function_list: Optional[List[str]] = None, # "search", "visit", "python" defult can be use without api key
            instruction: str = "",
            api_key: Optional[str] = None,
            base_url: Optional[str] = None,
            model: Optional[str] = None,
            use_xml_protocol: bool = True,  # XML protocol works better for IterResearch paradigm
    ):
        llm_config = dict(llm_config or {})
        if api_key:
            llm_config["api_key"] = api_key
        if base_url:
            llm_config["base_url"] = base_url

        self.llm_config = llm_config
        self.llm_generate_cfg = self.llm_config.get("generate_cfg", {})
        self.model = model or self.llm_config.get("model", LLM_MODEL_NAME)
        self.api_key = self.llm_config.get("api_key", LLM_API_KEY)
        self.base_url = self.llm_config.get("base_url", LLM_BASE_URL)
        self.max_input_tokens = self.llm_config.get("max_input_tokens", 32000)
        self.llm_timeout = self.llm_config.get("llm_timeout", 300.0)
        self.agent_timeout = self.llm_config.get("agent_timeout", 1800.0)
        self.function_list = function_list or list(TOOL_MAP.keys())
        self.instruction = instruction
        self.use_xml_protocol = use_xml_protocol

    def _get_tool_definitions(self) -> List[Dict]:
        """Get tool definitions in OpenAI function calling format."""
        tools = []
        for tool_name in self.function_list:
            if tool_name in TOOL_DESCRIPTIONS:
                tools.append(TOOL_DESCRIPTIONS[tool_name])
            else:
                # Fallback for custom tools
                tools.append({
                    "type": "function",
                    "function": {
                        "name": tool_name,
                        "description": f"Custom tool '{tool_name}'",
                        "parameters": {"type": "object", "properties": {}, "required": []}
                    }
                })
        return tools

    def parse_output(self, text: str) -> Dict[str, str]:
        """
        解析 LLM 的单次输出，严格提取 <plan>, <report>, 和 (<tool_call> 或 <answer> 或 <terminate>)。
        这是 IterResearch 范式的核心：LLM 在单次调用中生成三段式输出。
        """
        output = {
            "plan": "",
            "report": "",
            "tool_call": "",
            "answer": "",
            "terminate": False,
            "terminate_reason": "",
        }

        def _extract_last_block(pattern: str) -> str:
            matches = re.findall(pattern, text, flags=re.DOTALL | re.MULTILINE)
            for m in reversed(matches):
                if m and m.strip():
                    return m.strip()
            return ""

        # 1. 提取 <plan>
        plan = _extract_last_block(r"^\s*<plan>(.*?)</plan>")
        if plan:
            output["plan"] = plan

        # 2. 提取 <report>
        output["report"] = _extract_last_block(r"^\s*<report>(.*?)</report>")

        # 3. 提取 <tool_call>、<answer>、<terminate>
        output["tool_call"] = _extract_last_block(r"^\s*<tool_call>(.*?)</tool_call>")
        output["answer"] = _extract_last_block(r"^\s*<answer>(.*?)</answer>")
        term_body = _extract_last_block(r"^\s*<terminate>(.*?)</terminate>")
        if term_body != "":
            output["terminate"] = True
            output["terminate_reason"] = term_body

        if not output["tool_call"] and not output["answer"] and not output["terminate"]:
            logger.warning("LLM output did not contain <tool_call>, <answer>, or <terminate> tag.")

        return output

    async def call_server(
            self, 
            msgs: List[Dict], 
            stop_sequences: Optional[List[str]] = None,
            max_tries: int = 3,
            tools: Optional[List[Dict]] = None,
    ) -> Dict[str, Any]:
        """
        Call LLM server with support for both XML-based and native function calling.
        Uses AsyncOpenAI for true async concurrency.
        
        Returns:
            Dict with keys:
            - content: str, the text content
            - reasoning_content: Optional[str], thinking process (for models like DeepSeek R1)
            - tool_calls: Optional[List], native tool calls (when use_xml_protocol=False)
            - raw_message: the original message object
        """
        client = AsyncOpenAI(
            api_key=self.api_key or "EMPTY",
            base_url=self.base_url,
            timeout=self.llm_timeout,
        )

        base_sleep_time = 1
        stop_sequences = stop_sequences or (["<tool_response>"] if self.use_xml_protocol else None)

        for attempt in range(max_tries):
            try:
                request_params = {
                    "model": self.model,
                    "messages": msgs,
                    "temperature": self.llm_generate_cfg.get('temperature', 0.6),
                    "top_p": self.llm_generate_cfg.get('top_p', 0.95),
                    "presence_penalty": self.llm_generate_cfg.get('presence_penalty', 1.1)
                }
                
                # Add stop sequences only for XML mode
                if stop_sequences and self.use_xml_protocol:
                    request_params["stop"] = stop_sequences
                
                # Add tools for function calling mode (non-XML)
                if tools and not self.use_xml_protocol:
                    request_params["tools"] = tools
                
                # Add extra_body for thinking mode (DeepSeek R1 etc.)
                model_thinking_type = self.llm_generate_cfg.get("model_thinking_type", "")
                if model_thinking_type:
                    request_params["extra_body"] = {
                        "thinking": {"type": model_thinking_type}
                    }
                
                # Use native async call
                chat_response = await client.chat.completions.create(**request_params)

                message = chat_response.choices[0].message
                content = message.content or ""
                
                # Extract reasoning_content if available (DeepSeek R1, etc.)
                reasoning_content = getattr(message, 'reasoning_content', None)
                
                # Extract native tool_calls if available
                tool_calls = getattr(message, 'tool_calls', None)
                
                logger.debug(
                    f"Input messages: {msgs}, \n"
                    f"Reasoning_content: {reasoning_content}, \n"
                    f"Tool_calls: {tool_calls}, \n"
                    f"LLM Response: {content}"
                )
                
                return {
                    "content": content.strip() if content else "",
                    "reasoning_content": reasoning_content,
                    "tool_calls": tool_calls,
                    "raw_message": message,
                }

            except RateLimitError as e:
                logger.warning(f"Attempt {attempt + 1} rate limit error: {e}")
            except AuthenticationError as e:
                logger.error(f"Authentication error: {e}")
                break  # Don't retry auth errors
            except (APIError, APIConnectionError, APITimeoutError) as e:
                logger.warning(
                    f"Attempt {attempt + 1} API error: {e}, base_url: {self.base_url}, model: {self.model}")
            except Exception as e:
                logger.error(f"Attempt {attempt + 1} unexpected error: {e}")

            if attempt < max_tries - 1:
                sleep_time = base_sleep_time * (2 ** attempt) + random.uniform(0, 1)
                sleep_time = min(sleep_time, 30)
                logger.warning(f"Retrying in {sleep_time:.2f}s...")
                await asyncio.sleep(sleep_time)
            else:
                logger.error("All retry attempts exhausted. The LLM call failed.")
        
        return {
            "content": "LLM server error.",
            "reasoning_content": None,
            "tool_calls": None,
            "raw_message": None,
        }

    def count_tokens(self, messages, model=None):
        """Count tokens in messages"""
        if model is None:
            model = self.model or "gpt-4o"  # 使用实例的 model 或默认值
        try:
            # Convert dict messages to Message objects if needed
            full_message = []
            for x in messages:
                if isinstance(x, dict):
                    full_message.append(Message(**x))
                elif isinstance(x, Message):
                    full_message.append(x)
                else:
                    full_message.append(x)

            full_prompt = build_text_completion_prompt(full_message, allow_special=True)
            return count_tokens_base(full_prompt, model)
        except Exception as e:
            logger.warning(f"Failed to count tokens: {e}. Using simple split.")
            return sum(len(str(x).split()) for x in messages)

    async def _execute_function_call(self, tool_call) -> str:
        """Execute an OpenAI-style function call."""
        func_name = tool_call.function.name
        args_str = tool_call.function.arguments
        
        logger.debug(f"Function call: {func_name}({args_str})")
        
        if func_name not in TOOL_MAP:
            return f"Error: Tool {func_name} not found"
        
        try:
            args = json.loads(args_str) if args_str else {}
        except json.JSONDecodeError:
            return f"Error: Failed to decode arguments: {args_str}"
        
        tool = TOOL_MAP[func_name]
        loop = asyncio.get_event_loop()
        
        try:
            # Special handling for python tool: extract code from arguments
            if func_name == "python":
                code = args.get("code", "")
                if not code:
                    return "[Python Interpreter Error]: Empty code. Please provide code in arguments.code"
                result = await loop.run_in_executor(None, tool.call, code)
                return result if isinstance(result, str) else str(result)
            
            if asyncio.iscoroutinefunction(tool.call):
                if func_name == "parse_file":
                    params = {"files": args.get("files")}
                    result = await tool.call(params, file_root_path=FILE_DIR)
                else:
                    result = await tool.call(args)
            else:
                result = await loop.run_in_executor(None, tool.call, args)
            return result if isinstance(result, str) else str(result)
        except Exception as e:
            logger.error(f"Tool execution failed: {e}")
            return f"Error: Tool execution failed. {e}"

    async def _execute_xml_tool(self, tool_call_str: str) -> str:
        """Execute a tool call from XML <tool_call> block."""
        loop = asyncio.get_event_loop()

        try:
            # Check for <code> tag (Python code in XML mode)
            if "<code>" in tool_call_str and "</code>" in tool_call_str:
                code_raw = tool_call_str.split("<code>", 1)[1].rsplit("</code>", 1)[0].strip()
                result = await loop.run_in_executor(None, TOOL_MAP['python'].call, code_raw)
                return result

            # Parse JSON tool call
            tool_call = json5.loads(tool_call_str)
            tool_name = tool_call.get('name', '')
            tool_args = tool_call.get('arguments', {})

            if tool_name not in TOOL_MAP:
                return f"Error: Tool {tool_name} not found"

            tool = TOOL_MAP[tool_name]

            # Special handling for python tool: extract code from arguments
            if tool_name == "python":
                code = tool_args.get("code", "")
                if not code:
                    return "[Python Interpreter Error]: Empty code. Please provide code in arguments.code"
                result = await loop.run_in_executor(None, tool.call, code)
                return str(result) if not isinstance(result, str) else result

            # Handle async/sync tools
            if asyncio.iscoroutinefunction(tool.call):
                if tool_name == "parse_file":
                    params = {"files": tool_args.get("files")}
                    result = await tool.call(params, file_root_path=FILE_DIR)
                else:
                    result = await tool.call(tool_args)
            else:
                result = await loop.run_in_executor(None, tool.call, tool_args)

            return str(result) if not isinstance(result, str) else result

        except Exception as e:
            logger.error(f"Tool call parsing or execution failed: {e}")
            return f"Error: Tool call failed. Input: {tool_call_str}. Error: {e}"

    async def run(self, question, progress_callback: Optional[Callable[[Dict[str, Any]], Any]] = None):
        """
        Execute research following the IterResearch paradigm.
        
        Supports two modes:
        - Function Calling (default): Uses OpenAI-style tools parameter
        - XML Protocol: Uses <tool_call> tags in prompts
        
        Loop:
        s_t = (Q, R_{i-1}, O_{i-1})
        LLM(s_t) -> (Plan_i, Report_i, Action_i)
        
        if Action_i == <answer>:
            STOP
        else:
            O_i = Tool(Action_i)
            s_{t+1} = (Q, R_i, O_i)
            LOOP
        """

        async def emit(event: Dict[str, Any]):
            if not callable(progress_callback):
                return
            event.setdefault("timestamp", datetime.datetime.utcnow().isoformat())
            try:
                maybe = progress_callback(event)
                if inspect.isawaitable(maybe):
                    await maybe
            except Exception as callback_err:
                logger.warning(f"progress_callback raised error: {callback_err}")

        start_time = time.time()

        # Initialize research round
        research_round = ResearchRound(question=question)
        
        # Build system prompt based on mode (XML mode uses detailed prompt, function calling uses simpler one)
        if self.use_xml_protocol:
            system_prompt = get_iterresearch_system_prompt(
                today_date(), self.function_list, self.instruction, question=question
            )
        else:
            system_prompt = get_iterresearch_system_prompt_fc(
                today_date(), self.instruction, question=question
            )
        
        # Get tool definitions for function calling mode
        tool_definitions = self._get_tool_definitions() if not self.use_xml_protocol else None

        # Trajectory log for debugging
        full_trajectory_log = []
        prediction = ''
        termination = ''

        num_llm_calls_available = MAX_LLM_CALL_PER_RUN
        round_num = 0

        while num_llm_calls_available > 0:
            if time.time() - start_time > self.agent_timeout:
                logger.warning("Agent timeout reached.")
                termination = "timeout"
                prediction = "No answer found (timeout)."
                break

            round_num += 1
            num_llm_calls_available -= 1

            # Build prompt (s_t = Q, R_{i-1}, O_{i-1})
            current_context = research_round.get_context(system_prompt)

            if round_num == 1:
                full_trajectory_log.extend(current_context)

            # Single LLM call (generate P_i, R_i, A_i)
            request_msgs = current_context
            is_last_call = (num_llm_calls_available == 0)
            
            try:
                logger.debug(f"Round {round_num}: Calling LLM. Remaining calls: {num_llm_calls_available}")

                # Force final answer on last call
                if is_last_call:
                    finalize_instruction = (
                        "You have reached the maximum allowed LLM calls for this run. "
                        "Do not call tools anymore. Based on your current report and the information gathered so far, "
                        "provide the final answer now in the three-part format: "
                        "<plan>...</plan> <report>...</report> <answer>...</answer>"
                    )
                    request_msgs = current_context + [{"role": "user", "content": finalize_instruction}]

                response = await self.call_server(request_msgs, tools=tool_definitions)
                content = response["content"]
                reasoning_content = response["reasoning_content"]
                tool_calls = response["tool_calls"]
                raw_message = response["raw_message"]

                # Log reasoning content if available
                if reasoning_content:
                    logger.info(f"[Round {round_num}] Thinking: {reasoning_content[:500]}...")
                    await emit({"type": "thinking", "round": round_num, "content": reasoning_content})

                # Build assistant message for trajectory (preserving reasoning_content)
                if raw_message:
                    msg_dict = raw_message.model_dump(exclude_none=True)
                    if reasoning_content:
                        msg_dict['reasoning_content'] = reasoning_content
                    full_trajectory_log.append(msg_dict)
                else:
                    full_trajectory_log.append({"role": "assistant", "content": content})
                
                logger.debug(f'Round {round_num} LLM response received.')
            except Exception as e:
                logger.error(f"Unknown Error: {e}")
                prediction = f"Error: Unknown {e}"
                termination = 'unknown error'
                break

            # === Function Calling Mode: Handle native tool calls ===
            if not self.use_xml_protocol and tool_calls:
                # Execute each tool call
                tool_results = []
                for tool_call in tool_calls:
                    tool_result = await self._execute_function_call(tool_call)
                    logger.debug(f"Tool {tool_call.function.name} result: {tool_result[:200]}...")
                    
                    await emit({
                        "type": "tool",
                        "round": round_num,
                        "tool_name": tool_call.function.name,
                        "tool_args": tool_call.function.arguments,
                        "observation": tool_result[:1000],
                    })
                    tool_results.append(tool_result)
                
                # Update observation with combined tool results
                research_round.last_observation = "\n\n".join(tool_results)
                
                # Parse any report/answer from content (LLM may still output structured response)
                if content:
                    parsed = self.parse_output(content)
                    if parsed["report"]:
                        research_round.current_report = parsed["report"]
                    if parsed["answer"]:
                        prediction = parsed["answer"]
                        termination = 'answer found'
                        await emit({
                            "type": "final",
                            "round": round_num,
                            "answer": prediction,
                            "report": research_round.current_report,
                            "termination": termination,
                        })
                        break
                continue

            # === XML Protocol Mode: Parse structured output ===
            parsed = self.parse_output(content)
            plan_content = parsed["plan"]
            report_content = parsed["report"]
            action_content = parsed["tool_call"]
            answer_content = parsed["answer"]
            terminate_flag = parsed.get("terminate", False)
            terminate_reason = parsed.get("terminate_reason", "").strip()

            logger.debug(f"Round {round_num} - Plan: {plan_content}")
            logger.debug(f"Round {round_num} - Report: {report_content}")
            logger.debug(f"Round {round_num} - Action: {action_content}")
            logger.debug(f"Round {round_num} - Answer: {answer_content}")
            if terminate_flag:
                logger.debug(f"Round {round_num} - Terminate signaled. Reason: {terminate_reason}")

            await emit({
                "type": "round",
                "round": round_num,
                "plan": plan_content,
                "report": report_content,
                "action": action_content,
                "answer": answer_content,
                "terminate": terminate_flag,
                "reasoning_content": reasoning_content,
            })

            # Update report (R_i)
            if report_content:
                research_round.current_report = report_content
            else:
                logger.warning("No <report> found. Report will not be updated for the next round.")

            # Check for termination
            if answer_content:
                prediction = answer_content
                termination = 'answer found'
                if terminate_flag:
                    termination = 'terminate with answer'
                await emit({
                    "type": "final",
                    "round": round_num,
                    "answer": prediction,
                    "report": research_round.current_report,
                    "termination": termination,
                })
                logger.debug(f"Round {round_num}: LLM provided <answer>. Terminating loop.")
                break
            
            if terminate_flag:
                prediction = terminate_reason if terminate_reason else research_round.current_report.strip()
                termination = 'terminated by llm'
                await emit({
                    "type": "final",
                    "round": round_num,
                    "answer": prediction,
                    "report": research_round.current_report,
                    "termination": termination,
                })
                logger.debug(f"Round {round_num}: LLM signaled <terminate>. Final response prepared.")
                break
            
            if is_last_call:
                fallback_report = research_round.current_report.strip()
                fallback_source = fallback_report or terminate_reason
                prediction = fallback_source if fallback_source else research_round.last_observation
                termination = 'finalized without answer tag'
                await emit({
                    "type": "final",
                    "round": round_num,
                    "answer": prediction,
                    "report": research_round.current_report,
                    "termination": termination,
                })
                logger.warning("Last LLM call did not return <answer> or <terminate>. Promoting accumulated content as final answer.")
                break

            # Execute Action (A_i)
            if action_content:
                try:
                    logger.debug(f"Round {round_num}: Executing tool...")
                    tool_response_str = await self._execute_xml_tool(action_content)

                    # Store tool response O_i for next round s_{t+1}
                    research_round.last_observation = tool_response_str
                    await emit({
                        "type": "tool",
                        "round": round_num,
                        "tool_call": action_content,
                        "observation": tool_response_str,
                    })

                    # Log tool response
                    tool_obs_msg = f"{OBS_START}\n{tool_response_str}\n{OBS_END}"
                    full_trajectory_log.append({"role": "user", "content": tool_obs_msg})

                    logger.debug(f"Round {round_num}: Tool execution completed.")

                except Exception as e:
                    logger.error(f"Error calling tool: {e}")
                    error_str = f"Error executing tool: {e}"
                    research_round.last_observation = error_str
                    await emit({
                        "type": "tool_error",
                        "round": round_num,
                        "tool_call": action_content,
                        "observation": error_str,
                    })
                    full_trajectory_log.append({"role": "user", "content": f"{OBS_START}\n{error_str}\n{OBS_END}"})
            else:
                # LLM produced neither answer nor tool call
                logger.warning("LLM did not produce <answer> or <tool_call>. Forcing answer generation...")

                force_answer_msgs = current_context + [
                    {"role": "user", "content": (
                        "You did not provide a valid response format. "
                        "Based on your current report and the information gathered so far, "
                        "please provide the final answer to the original question. "
                        "Use the three-part format: <plan>...</plan> <report>...</report> <answer>...</answer>"
                    )}
                ]

                try:
                    forced_response = await self.call_server(force_answer_msgs, tools=None)
                    forced_content = forced_response["content"]
                    forced_parsed = self.parse_output(forced_content)

                    if forced_parsed["answer"]:
                        prediction = forced_parsed["answer"]
                        termination = "answer (forced)"
                        await emit({
                            "type": "final",
                            "round": round_num,
                            "answer": prediction,
                            "report": research_round.current_report,
                            "termination": termination,
                        })
                        full_trajectory_log.append({"role": "assistant", "content": forced_content})
                        break
                    else:
                        logger.error("Failed to force answer generation")
                        prediction = "No answer found (format error after retry)."
                        termination = "format error"
                        break
                except Exception as e:
                    logger.error(f"Error during forced answer generation: {e}")
                    prediction = "No answer found (format error)."
                    termination = "format error"
                    break

            # Token limit check
            token_count = self.count_tokens(request_msgs)
            logger.debug(f"Round {round_num} context token count: {token_count}")
            if token_count > self.max_input_tokens:
                logger.warning(f"Token quantity exceeds the limit: {token_count}")
                force_answer_msgs = current_context + [
                    {"role": "user", "content": "You have now reached the maximum context length. "
                                                "Stop making tool calls. Based on your research report, "
                                                "provide the final answer in the three-part format: "
                                                "<plan>...</plan> <report>...</report> <answer>...</answer>"}
                ]
                forced_response = await self.call_server(force_answer_msgs, tools=None)
                content = forced_response["content"]
                parsed = self.parse_output(content)
                prediction = parsed["answer"] if parsed["answer"] else "No answer found (token limit)."
                termination = 'token limit reached'
                await emit({
                    "type": "final",
                    "round": round_num,
                    "answer": prediction,
                    "report": research_round.current_report,
                    "termination": termination,
                })
                full_trajectory_log.append({"role": "assistant", "content": content})
                break

        # Finalize
        if not prediction:
            fallback_report = research_round.current_report.strip()
            if fallback_report:
                prediction = fallback_report
                if not termination:
                    termination = 'report fallback'
            elif num_llm_calls_available == 0:
                prediction = 'No answer found (exceeded available LLM calls).'
                termination = 'exceed available llm calls'
            else:
                prediction = 'No answer found.'
                termination = 'answer not found'

        # Return final result
        result = {
            "question": question,
            "prediction": prediction,
            "report": research_round.current_report,
            "termination": termination,
            "trajectory": full_trajectory_log,
        }

        await emit({
            "type": "status",
            "status": termination or 'completed',
            "answer": prediction,
            "report": research_round.current_report,
        })

        return result


async def main():
    # 1. 定义你的 LLM 配置
    llm_config = {
        "model": LLM_MODEL_NAME,
        "generate_cfg": {
            'max_input_tokens': 32000,
            "temperature": 0.6,
            "top_p": 0.95,
            "presence_penalty": 1.1
        }
    }

    # 2. 实例化 agent
    agent = WebResearcherAgent(
        llm_config=llm_config,
        function_list=['search', 'python']
    )

    question = '刘翔破纪录时候是多少岁?'
    gt = "23岁差2天"

    # 4. 运行完整的并行研究与总结
    final_result = await agent.run(question)

    # 5. 打印结果
    print(f"final_result: {final_result}")


if __name__ == "__main__":
    asyncio.run(main())



# -*- coding: utf-8 -*-
"""
Settings for WebResearcher Web UI
"""
from pathlib import Path
from typing import Optional

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """应用配置"""
    
    # 应用基础配置
    app_name: str = "WebResearcher Web UI"
    app_version: str = "0.2.0"
    debug: bool = False
    
    # 服务器配置
    host: str = "0.0.0.0"
    port: int = 8000
    
    # 数据目录
    data_dir: Optional[Path] = None
    
    # CORS 配置
    allow_origins: list = ["*"]
    
    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"


settings = Settings()



import json
from pathlib import Path
from typing import Any, AsyncGenerator, Dict

from fastapi import APIRouter, HTTPException
from fastapi.responses import StreamingResponse

from .session_manager import SessionManager
from .models import CreateSessionRequest, ResearchRequest

# 历史文件路径
HISTORY_FILE = Path(__file__).resolve().parent.parent / "data" / "history.jsonl"

# 创建路由
api_router = APIRouter(prefix="/api", tags=["research"])

# 创建会话管理器
manager = SessionManager(history_path=HISTORY_FILE)


def _format_sse(payload: Dict[str, Any]) -> str:
    return f"data: {json.dumps(payload, ensure_ascii=False)}\n\n"


@api_router.post("/session")
async def create_session(request: CreateSessionRequest):
    """创建新会话"""
    session = manager.create_session(instruction=request.instruction, tools=request.tools)
    return session.summary()


@api_router.post("/research")
async def submit_question(request: ResearchRequest):
    """在会话中提交新问题"""
    session = manager.get_session(request.session_id)
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")
    
    manager.start_research(session, request.question)
    return {"session_id": session.id, "status": "running"}


@api_router.get("/session/{session_id}")
async def fetch_session(session_id: str):
    """获取会话详情"""
    session = manager.get_session(session_id)
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")
    return session.to_dict(include_events=True)


@api_router.get("/session/{session_id}/turn/{turn_index}/process")
async def fetch_turn_process(session_id: str, turn_index: int):
    """获取指定轮次的研究过程（通过轮次索引）"""
    session = manager.get_session(session_id)
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")
    
    if turn_index < 0 or turn_index >= len(session.turns):
        raise HTTPException(status_code=404, detail="Turn not found")
    
    turn = session.turns[turn_index]
    process_data = turn.get_process_data()
    
    return {
        "session_id": session_id,
        "turn_index": turn_index,
        "task_id": turn.task_id,
        "question": turn.question,
        "answer": turn.answer,
        "status": turn.status,
        "process": process_data,
    }


@api_router.get("/session/{session_id}/task/{task_id}/process")
async def fetch_task_process(session_id: str, task_id: str):
    """获取指定任务的研究过程（通过 task_id）"""
    session = manager.get_session(session_id)
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")
    
    # 查找对应的 turn
    turn = None
    turn_index = -1
    for idx, t in enumerate(session.turns):
        if t.task_id == task_id:
            turn = t
            turn_index = idx
            break
    
    if not turn:
        raise HTTPException(status_code=404, detail="Task not found")
    
    process_data = turn.get_process_data()
    
    return {
        "session_id": session_id,
        "turn_index": turn_index,
        "task_id": turn.task_id,
        "question": turn.question,
        "answer": turn.answer,
        "status": turn.status,
        "process": process_data,
    }


@api_router.get("/session/{session_id}/stream")
async def stream_session(session_id: str):
    """流式获取会话事件"""
    session = manager.get_session(session_id)
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")
    
    # 如果是历史会话（没有 current_turn），直接返回空流
    if not session.current_turn:
        async def empty_generator() -> AsyncGenerator[str, None]:
            # 历史会话已完成，不需要流式传输
            yield _format_sse({"type": "info", "message": "Historical session, no live events"})
            return
        return StreamingResponse(empty_generator(), media_type="text/event-stream")
    
    condition = session.condition()
    
    async def event_generator() -> AsyncGenerator[str, None]:
        sent_count = 0
        
        while True:
            # 只发送当前轮次的事件
            current_turn = session.current_turn
            if not current_turn:
                break
            
            # 动态获取当前轮次索引（每次都重新计算，确保准确）
            current_turn_index = len(session.turns) - 1
            
            current_events = current_turn.events
            
            # 发送未发送的事件
            while sent_count < len(current_events):
                event_copy = dict(current_events[sent_count])
                event_copy["turn_index"] = current_turn_index
                yield _format_sse(event_copy)
                sent_count += 1
            
            # 检查当前轮次是否完成
            if current_turn.status in ("completed", "failed"):
                # 提取最后一轮的 report（从 events 中查找）
                last_report = ""
                for event in reversed(current_turn.events):
                    if event.get("type") == "round" and event.get("report"):
                        last_report = event["report"]
                        break
                
                # 发送最终状态
                yield _format_sse({
                    "type": "turn_finished",
                    "turn_index": current_turn_index,
                    "status": current_turn.status,
                    "answer": current_turn.answer,
                    "report": last_report,
                    "error": current_turn.error,
                })
                break
            
            # 等待新事件
            async with condition:
                await condition.wait()
    
    return StreamingResponse(event_generator(), media_type="text/event-stream")


@api_router.get("/history")
async def list_history(limit: int = 20):
    """获取历史会话列表"""
    records = await manager.read_history(limit=limit)
    return {"items": records}



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: React Agent implementing the MultiTurn ReAct paradigm.

Supports two tool calling modes:
1. OpenAI Function Calling (default): Uses OpenAI-style tools parameter, works with OpenAI/DeepSeek/etc.
2. XML Protocol: Uses <tool_call> tags, compatible with all LLMs including local models
"""
from typing import Any, Callable, Dict, List, Optional
import asyncio
import datetime
import inspect
import json
import json5
import random
import time
import re

from openai import (
    AsyncOpenAI,
    APIError,
    APIConnectionError,
    APITimeoutError,
    RateLimitError,
    AuthenticationError,
)

from webresearcher.base import today_date, Message, build_text_completion_prompt, count_tokens as count_tokens_base
from webresearcher.log import logger
from webresearcher.prompt import get_react_system_prompt_xml, TOOL_DESCRIPTIONS, get_react_system_prompt_fc
from webresearcher.tool_file import FileParser
from webresearcher.tool_scholar import Scholar
from webresearcher.tool_python import PythonInterpreter
from webresearcher.tool_search import Search
from webresearcher.tool_visit import Visit
from webresearcher.config import (
    LLM_API_KEY,
    LLM_BASE_URL,
    OBS_START,
    OBS_END,
    MAX_LLM_CALL_PER_RUN,
    FILE_DIR,
    LLM_MODEL_NAME,
)


TOOL_CLASS = [
    FileParser(),
    Scholar(),
    Visit(),
    Search(),
    PythonInterpreter(),
]
TOOL_MAP = {tool.name: tool for tool in TOOL_CLASS}


class ReactAgent:
    """
    A lightweight MultiTurn ReAct-style agent compatible with local vLLM or OpenAI endpoints.
    
    Supports two tool calling modes:
    - use_xml_protocol=False (default): OpenAI-style function calling, works with OpenAI/DeepSeek/etc.
    - use_xml_protocol=True: XML-based <tool_call> tags, compatible with all LLMs including local models
    """

    def __init__(
        self,
        llm_config: Optional[Dict] = None,
        function_list: Optional[List[str]] = None, # "search", "visit", "python" defult can be use without api key
        instruction: str = "",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        model: Optional[str] = None,
        use_xml_protocol: bool = False,
    ) -> None:
        llm_config = dict(llm_config or {})
        if api_key:
            llm_config["api_key"] = api_key
        if base_url:
            llm_config["base_url"] = base_url

        self.llm_config = llm_config
        self.model = model or self.llm_config.get("model", LLM_MODEL_NAME)
        self.generate_cfg = self.llm_config.get("generate_cfg", {"temperature": 0.6, "top_p": 0.95})
        self.api_key = self.llm_config.get("api_key", LLM_API_KEY)
        self.base_url = self.llm_config.get("base_url", LLM_BASE_URL)
        self.llm_timeout = self.llm_config.get("llm_timeout", 600.0)
        self.agent_timeout = self.llm_config.get("agent_timeout", 1800.0)

        self.function_list = function_list or list(TOOL_MAP.keys())
        self.instruction = instruction
        self.use_xml_protocol = use_xml_protocol

    def _get_tool_definitions(self) -> List[Dict]:
        """Get tool definitions in OpenAI function calling format."""
        tools = []
        for tool_name in self.function_list:
            if tool_name in TOOL_DESCRIPTIONS:
                tools.append(TOOL_DESCRIPTIONS[tool_name])
            else:
                # Fallback for custom tools
                tools.append({
                    "type": "function",
                    "function": {
                        "name": tool_name,
                        "description": f"Custom tool '{tool_name}'",
                        "parameters": {"type": "object", "properties": {}, "required": []}
                    }
                })
        return tools

    def count_tokens(self, messages: List[Dict]) -> int:
        try:
            full_message: List[Message] = []
            for x in messages:
                if isinstance(x, dict):
                    full_message.append(Message(**x))
                else:
                    full_message.append(x)
            full_prompt = build_text_completion_prompt(full_message, allow_special=True)
            return count_tokens_base(full_prompt, self.model)
        except Exception as e:
            logger.warning(f"Failed to count tokens: {e}. Using simple split.")
            return sum(len(str(x).split()) for x in messages)

    async def call_server(
        self, 
        msgs: List[Dict], 
        stop_sequences: Optional[List[str]] = None, 
        max_tries: int = 5,
        tools: Optional[List[Dict]] = None,
    ) -> Dict[str, Any]:
        """
        Call LLM server with support for both XML-based and native function calling.
        Uses AsyncOpenAI for true async concurrency.
        
        Returns:
            Dict with keys:
            - content: str, the text content
            - reasoning_content: Optional[str], thinking process (for models like DeepSeek)
            - tool_calls: Optional[List], native tool calls (when use_native_tools=True)
            - raw_message: the original message object
        """
        client = AsyncOpenAI(
            api_key=self.api_key or "EMPTY",
            base_url=self.base_url,
            timeout=self.llm_timeout,
        )
        base_sleep_time = 1
        stop_sequences = stop_sequences or ([OBS_START] if self.use_xml_protocol else None)

        for attempt in range(max_tries):
            try:
                request_params = {
                    "model": self.model,
                    "messages": msgs,
                    "temperature": self.generate_cfg.get("temperature", 0.6),
                    "top_p": self.generate_cfg.get("top_p", 0.95),
                }
                
                # Add stop sequences only for XML mode
                if stop_sequences and self.use_xml_protocol:
                    request_params["stop"] = stop_sequences
                
                # Add tools for function calling mode (non-XML)
                if tools and not self.use_xml_protocol:
                    request_params["tools"] = tools
                
                # Add extra_body for thinking mode (DeepSeek R1 etc.)
                model_thinking_type = self.generate_cfg.get("model_thinking_type", "")
                if model_thinking_type:
                    request_params["extra_body"] = {
                        "thinking": {"type": model_thinking_type}
                    }
                
                # Use native async call
                chat_response = await client.chat.completions.create(**request_params)
                
                message = chat_response.choices[0].message
                content = message.content or ""
                
                # Extract reasoning_content if available (DeepSeek R1, etc.)
                reasoning_content = getattr(message, 'reasoning_content', None)
                
                # Extract native tool_calls if available
                tool_calls = getattr(message, 'tool_calls', None)
                
                logger.debug(
                    f"Input messages: {msgs}, \n"
                    f"Reasoning_content: {reasoning_content}, \n"
                    f"Tool_calls: {tool_calls}, \n"
                    f"LLM Response: {content}"
                )
                
                return {
                    "content": content.strip() if content else "",
                    "reasoning_content": reasoning_content,
                    "tool_calls": tool_calls,
                    "raw_message": message,
                }
                
            except RateLimitError as e:
                logger.warning(f"Attempt {attempt + 1} rate limit error: {e}")
            except AuthenticationError as e:
                logger.error(f"Authentication error: {e}")
                break  # Don't retry auth errors
            except (APIError, APIConnectionError, APITimeoutError) as e:
                logger.warning(f"Attempt {attempt + 1} API error: {e}")
            except Exception as e:
                logger.error(f"Attempt {attempt + 1} unexpected error: {e}")

            if attempt < max_tries - 1:
                sleep_time = base_sleep_time * (2 ** attempt) + random.uniform(0, 1)
                sleep_time = min(sleep_time, 30)
                logger.warning(f"Retrying in {sleep_time:.2f}s...")
                await asyncio.sleep(sleep_time)
        
        return {
            "content": "LLM server error.",
            "reasoning_content": None,
            "tool_calls": None,
            "raw_message": None,
        }

    @staticmethod
    def _strip_after_tool_response(content: str) -> str:
        if OBS_START in content:
            return content.split(OBS_START, 1)[0].strip()
        return content

    async def _execute_function_call(self, tool_call) -> str:
        """Execute an OpenAI-style function call."""
        func_name = tool_call.function.name
        args_str = tool_call.function.arguments
        
        logger.debug(f"Native tool call: {func_name}({args_str})")
        
        if func_name not in TOOL_MAP:
            return f"Error: Tool {func_name} not found"
        
        try:
            args = json.loads(args_str) if args_str else {}
        except json.JSONDecodeError:
            return f"Error: Failed to decode arguments: {args_str}"
        
        tool = TOOL_MAP[func_name]
        try:
            # Special handling for python tool: extract code from arguments
            if func_name == "python":
                code = args.get("code", "")
                if not code:
                    return "[Python Interpreter Error]: Empty code. Please provide code in arguments.code"
                loop = asyncio.get_event_loop()
                result = await loop.run_in_executor(None, tool.call, code)
                return result if isinstance(result, str) else str(result)
            
            if asyncio.iscoroutinefunction(tool.call):
                if func_name == "parse_file":
                    params = {"files": args.get("files")}
                    result = await tool.call(params, file_root_path=FILE_DIR)
                else:
                    result = await tool.call(args)
            else:
                loop = asyncio.get_event_loop()
                result = await loop.run_in_executor(None, tool.call, args)
            return result if isinstance(result, str) else str(result)
        except Exception as e:
            logger.error(f"Tool execution failed: {e}")
            return f"Error: Tool execution failed. {e}"

    async def _execute_xml_tool(self, tool_call_block: str) -> str:
        """Execute a tool call from XML <tool_call> block."""
        # Python inline code path
        if "<code>" in tool_call_block and "</code>" in tool_call_block and "python" in tool_call_block.lower():
            code_raw = tool_call_block.split("<code>", 1)[1].split("</code>", 1)[0].strip()
            result = TOOL_MAP["python"].call(code_raw)
            return result if isinstance(result, str) else str(result)

        # JSON tool path
        try:
            tool_call = json5.loads(tool_call_block)
            tool_name = tool_call.get("name", "")
            tool_args = tool_call.get("arguments", {})
        except Exception:
            return 'Error: Tool call is not a valid JSON. Tool call must contain a valid "name" and "arguments" field.'

        if tool_name not in TOOL_MAP:
            return f"Error: Tool {tool_name} not found"

        tool = TOOL_MAP[tool_name]
        # handle async tool (parse_file) with file root
        try:
            if asyncio.iscoroutinefunction(tool.call):
                if tool_name == "parse_file":
                    params = {"files": tool_args.get("files")}
                    result = await tool.call(params, file_root_path=FILE_DIR)
                else:
                    result = await tool.call(tool_args)
            else:
                loop = asyncio.get_event_loop()
                result = await loop.run_in_executor(None, tool.call, tool_args)
            return result if isinstance(result, str) else str(result)
        except Exception as e:
            logger.error(f"Tool execution failed: {e}")
            return f"Error: Tool execution failed. {e}"

    def _parse_answer(self, content: str) -> Dict[str, Optional[str]]:
        ans = {
            "answer": None,
            "terminate": False,
        }
        # Prefer <answer> as a termination signal; if both exist, <answer> wins
        answer_match = re.search(r"<answer>(.*?)</answer>", content, re.DOTALL)
        if answer_match:
            ans["answer"] = answer_match.group(1).strip()
            ans["terminate"] = True  # Treat <answer> as a terminate signal
            return ans
        term_match = re.search(r"<terminate>(.*?)</terminate>", content, re.DOTALL)
        if term_match:
            ans["terminate"] = True
            body = term_match.group(1)
            if body:
                ans["answer"] = body.strip()
        return ans

    async def run(self, question: str, progress_callback: Optional[Callable[[Dict[str, Any]], Any]] = None) -> Dict[str, str]:
        """
        Run the agent loop.
        
        Args:
            question: User's question
            progress_callback: Optional callback for progress updates
            
        Returns:
            Dict with question, prediction, termination, and trajectory
        """
        async def emit(event: Dict[str, Any]):
            if not callable(progress_callback):
                return
            event.setdefault("timestamp", datetime.datetime.utcnow().isoformat())
            try:
                maybe = progress_callback(event)
                if inspect.isawaitable(maybe):
                    await maybe
            except Exception as callback_err:
                logger.warning(f"progress_callback raised error: {callback_err}")

        # Build system prompt (XML mode uses detailed prompt, function calling uses simpler one)
        if self.use_xml_protocol:
            system_prompt = get_react_system_prompt_xml(today_date(), self.function_list, self.instruction, question=question)
        else:
            system_prompt = get_react_system_prompt_fc(today_date(), self.function_list, self.instruction, question=question)
        
        messages: List[Dict] = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": question},
        ]
        
        # Get tool definitions for function calling mode
        tool_definitions = self._get_tool_definitions() if not self.use_xml_protocol else None

        start_time = time.time()
        remaining = MAX_LLM_CALL_PER_RUN
        round_num = 0

        while remaining > 0:
            if time.time() - start_time > self.agent_timeout:
                best_effort = "Final answer generated by agent (timeout)."
                await emit({"type": "final", "answer": best_effort, "termination": "timeout"})
                return {
                    "question": question,
                    "prediction": best_effort,
                    "termination": "timeout",
                    "trajectory": messages,
                }

            remaining -= 1
            round_num += 1
            
            response = await self.call_server(messages, tools=tool_definitions)
            content = response["content"]
            reasoning_content = response["reasoning_content"]
            tool_calls = response["tool_calls"]
            raw_message = response["raw_message"]
            
            # Log reasoning content if available
            if reasoning_content:
                logger.info(f"[Round {round_num}] Thinking: {reasoning_content[:500]}...")
                await emit({"type": "thinking", "round": round_num, "content": reasoning_content})
            
            # === OpenAI Function Calling Mode ===
            if not self.use_xml_protocol and tool_calls:
                # Append assistant message with tool_calls (preserving reasoning_content)
                msg_dict = raw_message.model_dump(exclude_none=True) if raw_message else {"role": "assistant", "content": content}
                if reasoning_content:
                    msg_dict['reasoning_content'] = reasoning_content
                messages.append(msg_dict)
                
                # Execute each tool call
                for tool_call in tool_calls:
                    tool_result = await self._execute_function_call(tool_call)
                    logger.debug(f"Tool {tool_call.function.name} result: {tool_result[:200]}...")
                    
                    await emit({
                        "type": "tool",
                        "round": round_num,
                        "tool_name": tool_call.function.name,
                        "tool_args": tool_call.function.arguments,
                        "observation": tool_result[:1000],
                    })
                    
                    # Append tool result
                    messages.append({
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "content": tool_result
                    })
                continue
            
            # === XML Protocol Mode ===
            content = self._strip_after_tool_response(content)
            
            if "<tool_call>" in content and "</tool_call>" in content:
                tool_block = content.split("<tool_call>", 1)[1].split("</tool_call>", 1)[0]
                tool_result = await self._execute_xml_tool(tool_block)
                
                await emit({
                    "type": "tool",
                    "round": round_num,
                    "tool_call": tool_block,
                    "observation": tool_result[:1000],
                })
                
                # Combine tool_call and response into a single 'user' message
                messages.append({
                    "role": "user",
                    "content": (
                        f"<tool_call>\n{tool_block}\n</tool_call>\n"
                        f"{OBS_START}\n{tool_result}\n{OBS_END}"
                    )
                })
                continue

            # === Normal Response Path (no tool call) ===
            messages.append({"role": "assistant", "content": content})

            # Check for termination
            final = self._parse_answer(content)
            logger.debug(f"Final answer check: {final}")
            
            if final["answer"]:
                await emit({"type": "final", "round": round_num, "answer": final["answer"], "termination": "terminated with answer"})
                return {
                    "question": question,
                    "prediction": final["answer"],
                    "termination": "terminated with answer",
                    "trajectory": messages,
                }
            
            if final["terminate"]:
                best_effort = content.strip() or "Final answer generated by agent."
                await emit({"type": "final", "round": round_num, "answer": best_effort, "termination": "terminated without answer"})
                return {
                    "question": question,
                    "prediction": best_effort,
                    "termination": "terminated without answer",
                    "trajectory": messages,
                }
            
            # For function calling mode without tool calls, treat content as final answer
            if not self.use_xml_protocol and not tool_calls:
                await emit({"type": "final", "round": round_num, "answer": content, "termination": "function calling completed"})
                return {
                    "question": question,
                    "prediction": content,
                    "termination": "function calling completed",
                    "trajectory": messages,
                }
            
            # Prompt to continue (XML protocol mode)
            messages.append({
                "role": "user",
                "content": "Please continue your analysis or provide the final answer using <answer> tags."
            })

            # Last round fallback
            if remaining == 0:
                forced_prompt = (
                    "You have reached the limit. Stop tool calls. Provide the final response using "
                    "<answer> only. Do NOT include <tool_call> or <think>."
                )
                if self.instruction:
                    forced_prompt = f"{forced_prompt}\n\nRemember the task-specific instruction:\n{self.instruction}"
                
                messages.append({"role": "user", "content": forced_prompt})
                response = await self.call_server(messages, tools=None)  # No tools for final call
                content = response["content"]
                messages.append({"role": "assistant", "content": content})
                
                final = self._parse_answer(content)
                if final["answer"]:
                    await emit({"type": "final", "round": round_num, "answer": final["answer"], "termination": "terminated with answer (forced)"})
                    return {
                        "question": question,
                        "prediction": final["answer"],
                        "termination": "terminated with answer (forced)",
                        "trajectory": messages,
                    }
                
                fallback_text = content.strip() or "Final answer generated by agent."
                await emit({"type": "final", "round": round_num, "answer": fallback_text, "termination": "finalized without answer tag"})
                return {
                    "question": question,
                    "prediction": fallback_text,
                    "termination": "finalized without answer tag",
                    "trajectory": messages,
                }

        # Exhausted LLM calls
        forced_prompt = (
            "You have reached the limit. Stop tool calls. Provide the final response using "
            "<answer> only. Do NOT include <tool_call> or <think>."
        )
        if self.instruction:
            forced_prompt = f"{forced_prompt}\n\nRemember the task-specific instruction:\n{self.instruction}"
        
        messages.append({"role": "user", "content": forced_prompt})
        response = await self.call_server(messages, tools=None)
        content = response["content"]
        messages.append({"role": "assistant", "content": content})
        
        final = self._parse_answer(content)
        if final["answer"]:
            await emit({"type": "final", "answer": final["answer"], "termination": "terminated with answer (forced)"})
            return {
                "question": question,
                "prediction": final["answer"],
                "termination": "terminated with answer (forced)",
                "trajectory": messages,
            }
        
        fallback_text = content.strip() or "Final answer generated by agent."
        await emit({"type": "final", "answer": fallback_text, "termination": "exceed available llm calls"})
        return {
            "question": question,
            "prediction": fallback_text,
            "termination": "exceed available llm calls (finalized without answer tag)",
            "trajectory": messages,
        }

async def main():
    agent = ReactAgent()
    result = await agent.run("What is the capital of France?")
    print(result)

if __name__ == "__main__":
    asyncio.run(main())


from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class CreateSessionRequest(BaseModel):
    """创建新会话"""
    instruction: Optional[str] = Field(default="", max_length=2000, description="可选的附加指令")
    tools: Optional[List[str]] = Field(default=None, description="限定可用工具名称列表")


class ResearchRequest(BaseModel):
    """在会话中提交新问题"""
    session_id: str = Field(..., description="会话 ID")
    question: str = Field(..., min_length=1, max_length=4000, description="研究问题")


class TaskEvent(BaseModel):
    """返回给前端的事件结构，用于绘制时间轴或日志"""

    type: str
    timestamp: datetime
    round: Optional[int] = None
    plan: Optional[str] = None
    report: Optional[str] = None
    action: Optional[str] = None
    tool_call: Optional[str] = None
    observation: Optional[str] = None
    answer: Optional[str] = None
    status: Optional[str] = None
    termination: Optional[str] = None


class TaskPayload(BaseModel):
    """任务详情响应模型"""

    task_id: str
    question: str
    status: str
    created_at: datetime
    updated_at: datetime
    events: List[Dict[str, Any]]
    result: Optional[Dict[str, Any]] = None
    error: Optional[str] = None
